"""
### Perform convolutions on arrays.  ###
"""

# Standard Library
from typing import List, Tuple, Optional, Union

# External
import numpy as np
from numba import jit, prange

# Internal
from buteo.utils import core_utils
from buteo.raster import convolution_funcs



@jit(nopython=True, nogil=True, cache=True, fastmath=True, inline='always')
def simple_blur_kernel_2d_3x3() -> Tuple[np.ndarray, np.ndarray]:
    """
    Create a 2D blur kernel.

    [1, 2, 1],
    [2, 4, 2],
    [1, 2, 1],

    Returns:
        Tuple[np.ndarray, np.ndarray]: The offsets and weights.
    """
    offsets = np.array([
        [ 1, -1], [ 1, 0], [ 1, 1],
        [ 0, -1], [ 0, 0], [ 0, 1],
        [-1, -1], [-1, 0], [-1, 1],
    ])

    weights = np.array([
        0.08422299, 0.12822174, 0.08422299,
        0.12822174, 0.1502211 , 0.12822174,
        0.08422299, 0.12822174, 0.08422299,
    ], dtype=np.float32)

    return offsets, weights


@jit(nopython=True, nogil=True, cache=True, fastmath=True, inline='always')
def simple_unsharp_kernel_2d_3x3(
    strength: float = 1.0
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Create a 2D unsharp kernel.

    baseweights:
        0.09911165, 0.15088834, 0.09911165,
        0.15088834, 0.        , 0.15088834,
        0.09911165, 0.15088834, 0.09911165,

    Returns:
        Tuple[np.ndarray, np.ndarray]: The offsets and weights.
    """
    offsets = np.array([
        [ 1, -1], [ 1, 0], [ 1, 1],
        [ 0, -1], [ 0, 0], [ 0, 1],
        [-1, -1], [-1, 0], [-1, 1],
    ])

    base_weights = np.array([
        0.09911165, 0.15088834, 0.09911165,
        0.15088834, 0.        , 0.15088834,
        0.09911165, 0.15088834, 0.09911165,
    ], dtype=np.float32)

    weights = base_weights * strength
    middle_weight = np.sum(weights) + 1.0
    weights *= -1.0
    weights[4] = middle_weight

    return offsets, weights


def simple_shift_kernel_2d(
    x_offset: float,
    y_offset: float,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Create a 2D shift kernel. Useful for either aligning rasters at the sub-pixel
    level or for shifting a raster by a whole pixel while keeping the bbox.
    Can be used to for an augmentation, where channel misalignment is simulated.

    Args:
        x_offset (float): The x offset.
        y_offset (float): The y offset.

    Returns:
        Tuple[np.ndarray, np.ndarray]: The offsets and weights.
    """
    if x_offset == 0.0 and y_offset == 0.0:
        offsets = np.array([[0, 0]], dtype=np.int64)
        weights = np.array([1.0], dtype=np.float32)

        return offsets, weights

    y0 = [int(np.floor(y_offset)), int(np.ceil(y_offset))] if y_offset != 0 else [0, 0]
    x0 = [int(np.floor(x_offset)), int(np.ceil(x_offset))] if x_offset != 0 else [0, 0]

    if x_offset == 0.0 or x_offset % 1 == 0.0:
        offsets = np.zeros((2, 2), dtype=np.int64)
        weights = np.zeros(2, dtype=np.float32)

        offsets[0] = [int(x_offset) if x_offset % 1 == 0.0 else 0, y0[0]]
        offsets[1] = [int(x_offset) if x_offset % 1 == 0.0 else 0, y0[1]]

        weights[0] = y_offset - y0[0]
        weights[1] = 1 - weights[0]

    elif y_offset == 0.0 or y_offset % 1 == 0.0:
        offsets = np.zeros((2, 2), dtype=np.int64)
        weights = np.zeros(2, dtype=np.float32)

        offsets[0] = [x0[0], int(y_offset) if y_offset % 1 == 0.0 else 0]
        offsets[1] = [x0[1], int(y_offset) if y_offset % 1 == 0.0 else 0]

        weights[0] = x_offset - x0[0]
        weights[1] = 1 - weights[0]

    else:
        offsets = np.zeros((4, 2), dtype=np.int64)
        weights = np.zeros(4, dtype=np.float32)

        offsets[0] = [x0[0], y0[0]]
        offsets[1] = [x0[0], y0[1]]
        offsets[2] = [x0[1], y0[0]]
        offsets[3] = [x0[1], y0[1]]

        weights[0] = (1 - (x_offset - offsets[0][0])) * (1 - (y_offset - offsets[0][1]))
        weights[1] = (1 - (x_offset - offsets[1][0])) * (1 + (y_offset - offsets[1][1]))
        weights[2] = (1 + (x_offset - offsets[2][0])) * (1 - (y_offset - offsets[2][1]))
        weights[3] = (1 + (x_offset - offsets[3][0])) * (1 + (y_offset - offsets[3][1]))

    return offsets, weights


def weight_distance(
    arr: np.ndarray,
    method: Optional[str] = None,
    decay: float = 0.2,
    sigma: float = 1.0,
    center: float = 0.0,
    spherical: bool = False,
    radius: float = 3.0,
) -> float:
    """
    Weights the kernel by distance using various methods.

    Args:
        arr (numpy.ndarray): The input array.
        method (str=None): The weighting method to use.
            "none": No weighting (default).
            "linear": Linear decay.
            "sqrt": Square root decay.
            "power": Power decay.
            "log": Logarithmic decay.
            "gaussian": Gaussian decay.
        decay (float=0.2): The decay rate for the `linear`, `sqrt`, and `power` methods.
        sigma (float=1.0): The standard deviation for the Gaussian method.
        center (float=0.0): The center of the array.
        spherical (bool=False): If True, adjust weights based on the radius.
        radius (float=3.0): The radius for spherical adjustments.

    Returns:
        float: The computed weight.
    """

    if center == 0.0:
        normed = np.linalg.norm(arr)
    else:
        normed = np.linalg.norm(arr - np.array([0, 0, center]))

    if normed == 0.0:
        weights = 1.0

    if method is None or method == "none" or method == "":
        weights = 1.0
    elif method == "linear":
        weights = np.power((1 - decay), normed)
    elif method == "sqrt":
        weights = np.power(np.sqrt((1 - decay)), normed)
    elif method == "power":
        weights = np.power(np.power((1 - decay), 2), normed)
    elif method == "log":
        weights = np.log(normed + 2) # +2 to avoid log(0)
    elif method == "gaussian":
        weights = np.exp(-(np.power(normed, 2)) / (2 * np.power(sigma, 2)))
    else:
        raise ValueError("Unable to parse parameters for weight_distance.")

    if spherical:
        sqrt_2 = np.sqrt(2)
        half_sqrt_2 = np.divide(sqrt_2, 2)

        if normed > radius + half_sqrt_2:
            return 0.0
        elif normed < radius - half_sqrt_2:
            return 1.0

        dist_min = radius - half_sqrt_2
        dist_max = radius + half_sqrt_2

        normed = 1 - np.interp(normed, [dist_min, dist_max], [0, 1])
        return weights * normed

    return weights


def rotate_kernel(bottom_right: np.ndarray) -> np.ndarray:
    """
    Creates a whole kernel from a quadrant.

    Args:
        bottom_right (numpy.ndarray): The bottom-right quadrant of the kernel.

    Returns:
        numpy.ndarray: The complete kernel generated from the given quadrant.
    """
    size = ((bottom_right.shape[0] - 1) * 2) + 1
    depth = bottom_right.shape[2]
    kernel = np.zeros((size, size, depth), dtype="float32")

    top_right = np.flipud(bottom_right)
    lower_left = np.fliplr(bottom_right)
    top_left = np.flipud(lower_left)

    kernel[size // 2:, size // 2:, :] = bottom_right
    kernel[:1 + -size // 2, :1 + -size // 2, :] = top_left
    kernel[1 + size // 2:, :size // 2, :] = lower_left[1:, :-1, :]
    kernel[:size // 2, 1 + size // 2:, :] = top_right[:-1, 1:, :]

    return kernel


def get_kernel(
    size: int,
    depth: int = 1,
    hole: bool = False,
    inverted: bool = False,
    normalise: bool = True,
    multi_dimensional: bool = False,
    multi_dimensional_center: int = 0,
    spherical: bool = False,
    distance_weight: Optional[str] = None,
    distance_decay: float = 0.2,
    distance_sigma: float = 1,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Generates a square kernel for convolutions.

    Args:
        size (int): Size of the kernel (must be odd).
        depth (int=1): Depth of the kernel.
        hole (bool=False): Create a hole in the center of the kernel.
        inverted (bool=False): Invert the kernel values.
        normalise (bool=True): Normalize the kernel values.
        multi_dimensional (bool=False): Consider the kernel multi-dimensional.
        multi_dimensional_center (int=0): Center of the
            multi-dimensional kernel.
        spherical (bool=False): Consider the kernel spherical.
        distance_weight (str or None=None): Distance weighting method.
        distance_decay (float=0.2): Distance decay factor.
        distance_sigma (float=1): Distance sigma for Gaussian distance weighting.

    Returns:
        tuple: A tuple containing the kernel, weights, and offsets.
    """
    assert size >= 3, "Kernel must have atleast size 3."
    assert size % 2 != 0, "Kernel must be an uneven size."
    assert isinstance(size, int), "Kernel must be an integer."
    assert depth >= 1, "Depth must be a positive integer"
    assert isinstance(depth, int), "Depth must be an integer."

    if distance_weight is False:
        distance_weight = None

    quadrant = np.zeros((1 + size // 2, 1 + size // 2, depth), dtype="float32")

    for idx_x in range(0, quadrant.shape[0]):
        for idx_y in range(0, quadrant.shape[1]):
            for idx_z in range(0, quadrant.shape[2]):

                z_value = idx_z if multi_dimensional else 0

                weighted = weight_distance(
                    np.array([idx_x, idx_y, z_value], dtype="float32"),
                    method=distance_weight,
                    decay=distance_decay,
                    sigma=distance_sigma,
                    center=multi_dimensional_center,
                    spherical=spherical,
                    radius=size / 2,
                )

                quadrant[idx_x, idx_y, idx_z] = weighted

    if hole:
        for idx_z in range(0, quadrant.shape[2]):
            quadrant[0, 0, idx_z] = 0

    kernel = rotate_kernel(quadrant)

    if distance_weight == "log":
        kernel = kernel.max() - kernel

    if inverted:
        kernel = 1 - kernel

    if normalise:
        if multi_dimensional:
            summed = kernel.sum()
            if summed != 0.0:
                kernel = kernel / summed
        else:
            summed = kernel.sum(axis=(0, 1))

            for dim in range(0, depth):
                kernel[:, :, dim] = kernel[:, :, dim] / summed[dim]

    weights = []
    offsets = []

    for idx_x in range(0, kernel.shape[0]):
        for idx_y in range(0, kernel.shape[1]):
            for idx_z in range(0, kernel.shape[2]):
                current_weight = kernel[idx_x][idx_y][idx_z]

                if current_weight <= 0.0:
                    continue

                offsets.append(
                    [
                        idx_x - (kernel.shape[0] // 2),
                        idx_y - (kernel.shape[1] // 2),
                        idx_z
                    ]
                )

                weights.append(current_weight)

    return kernel, np.array(weights, dtype="float32"), np.array(offsets, dtype=int)


def pad_array(
    arr: np.ndarray,
    pad_size: int = 1,
    method: str = "same",
    constant_value: Union[float, int] = 0.0,
) -> np.ndarray:
    """
    Create a padded view of an array using SAME padding.

    Args:
        arr (numpy.ndarray): The input array to be padded.

    Keyword Args:
        pad_size (int=1): The number of padding elements to add
            to each side of the array. Default is 1.
        method (str="same"): The padding method to use. Default
            is "same". Other options are "edge" and "constant".
        constant_value (int=None): The constant value to use
            when padding with "constant". Default is 0.

    Returns:
        numpy.ndarray: A padded view of the input array.
    """
    core_utils.type_check(arr, [np.ndarray], "arr")
    core_utils.type_check(pad_size, [int], "pad_size")
    core_utils.type_check(method, [str], "method")

    assert pad_size >= 0, "pad_size must be a non-negative integer"
    assert method in ["same", "SAME", "edge", "EDGE"], "method must be one of ['same', 'SAME', 'constant', 'CONSTANT']"

    if method in ["same", "SAME"]:
        padded_view = np.pad(
            arr,
            pad_width=((pad_size, pad_size), (pad_size, pad_size), (0, 0)),
            mode='edge',
        )
    elif method in ["constant", "CONSTANT"]:
        padded_view = np.pad(
            arr,
            pad_width=((pad_size, pad_size), (pad_size, pad_size), (0, 0)),
            mode='constant',
            constant_values=constant_value,
        )

    return padded_view

METHOD_ENUMS = {
    "sum": 1,
    "mode": 2,
    "max": 3,
    "dilate": 3,
    "min": 4,
    "erode": 4,
    "contrast": 5,
    "median": 6,
    "std": 7,
    "mad": 8,
    "z_score": 9,
    "z_score_mad": 10,
    "sigma_lee": 11,
    "quantile": 12,
    "occurrances": 13,
    "feather": 14,
    "roughness": 15,
    "roughness_tri": 16,
    "roughness_tpi": 17,
}


@jit(nopython=True, nogil=False, fastmath=True, cache=True)
def hood_to_value(method, values, weights, nodata_value=-9999.9, center_idx=0, value=0.5):
    """ Convert a array of values and weights to a single value using a given method. """
    if method == 1:
        return convolution_funcs.hood_sum(values, weights)
    elif method == 2:
        return convolution_funcs.hood_mode(values, weights)
    elif method == 3:
        return convolution_funcs.hood_max(values, weights)
    elif method == 4:
        return convolution_funcs.hood_min(values, weights)
    elif method == 5:
        return convolution_funcs.hood_contrast(values, weights)
    elif method == 6:
        return convolution_funcs.hood_quantile(values, weights, 0.5)
    elif method == 7:
        return convolution_funcs.hood_standard_deviation(values, weights)
    elif method == 8:
        return convolution_funcs.hood_median_absolute_deviation(values, weights)
    elif method == 9:
        return convolution_funcs.hood_z_score(values, weights, center_idx)
    elif method == 10:
        return convolution_funcs.hood_z_score_mad(values, weights, center_idx)
    elif method == 11:
        return convolution_funcs.hood_sigma_lee(values, weights)
    elif method == 12:
        return convolution_funcs.hood_quantile(values, weights, value)
    elif method == 13:
        return convolution_funcs.hood_count_occurances(values, weights, value, normalise=False)
    elif method == 14:
        return convolution_funcs.hood_count_occurances(values, weights, value, normalise=True)
    elif method == 15:
        return convolution_funcs.hood_roughness(values, weights, center_idx)
    elif method == 16:
        return convolution_funcs.hood_roughness_tri(values, weights, center_idx)
    elif method == 17:
        return convolution_funcs.hood_roughness_tpi(values, weights, center_idx)
    else:
        return nodata_value



@jit(nopython=True, parallel=True, nogil=False, fastmath=True, cache=True)
def _convolve_array_collapse(
    arr: np.ndarray,
    offsets: List[Tuple[int, int, int]],
    weights: List[float],
    method: int = 1,
    nodata: bool = False,
    nodata_value: float = -9999.9,
    normalise_edges: bool = True,
    value: Union[int, float] = 0.5,
) -> np.ndarray:
    """
    Internal. Convolve an array using a set of offsets and weights.
    """
    result = np.zeros((arr.shape[0], arr.shape[1], 1), dtype="float32")
    hood_size = len(offsets)

    for idx_y in prange(0, arr.shape[0]):
        for idx_x in range(0, arr.shape[1]):

            center_idx = 0

            if nodata and arr[idx_y, idx_x] == nodata_value:
                result[idx_y, idx_x] = nodata_value
                continue

            hood_normalise = False
            hood_values = np.zeros(hood_size, dtype="float32")
            hood_weights = np.zeros(hood_size, dtype="float32")
            hood_count = 0

            for idx_h in range(0, hood_size):
                hood_x = idx_x + offsets[idx_h][0]
                hood_y = idx_y + offsets[idx_h][1]
                hood_z = offsets[idx_h][2]

                if hood_x < 0 or hood_x >= arr.shape[1]:
                    if normalise_edges:
                        hood_normalise = True
                    continue

                if hood_y < 0 or hood_y >= arr.shape[0]:
                    if normalise_edges:
                        hood_normalise = True
                    continue

                if hood_z < 0 or hood_z >= arr.shape[2]:
                    if normalise_edges:
                        hood_normalise = True
                    continue

                if nodata and arr[hood_y, hood_x, hood_z] == nodata_value:
                    if normalise_edges:
                        hood_normalise = True
                    continue

                hood_values[hood_count] = arr[hood_y, hood_x, hood_z]
                hood_weights[hood_count] = weights[idx_h]
                hood_count += 1

                if offsets[idx_h][0] == 0 and offsets[idx_h][1] == 0:
                    center_idx = hood_count - 1

            if hood_count == 0:
                result[idx_y, idx_x] = nodata_value
                continue

            hood_values = hood_values[:hood_count]
            hood_weights = hood_weights[:hood_count]

            if hood_normalise:
                hood_weights /= np.sum(hood_weights)

            result[idx_y, idx_x, 0] = hood_to_value(method, hood_values, hood_weights, nodata_value, center_idx, value)

    return result


@jit(nopython=True, parallel=True, nogil=False, fastmath=True, cache=True)
def _convolve_array(
    arr: np.ndarray,
    offsets: List[Tuple[int, int, int]],
    weights: List[float],
    method: int = 1,
    nodata: bool = False,
    nodata_value: float = -9999.9,
    normalise_edges: bool = True,
    value: Union[int, float, None] = None,
) -> np.ndarray:
    """
    Internal. Convolve an array using a set of offsets and weights.
    """
    result = np.zeros((arr.shape[0], arr.shape[1], arr.shape[2]), dtype="float32")
    hood_size = len(offsets)

    for idx_y in prange(0, arr.shape[0]):
        for idx_x in range(0, arr.shape[1]):
            for idx_z in range(0, arr.shape[2]):

                center_idx = 0
                hood_normalise = False
                hood_values = np.zeros(hood_size, dtype="float32")
                hood_weights = np.zeros(hood_size, dtype="float32")
                hood_count = 0

                if nodata and arr[idx_y, idx_x, idx_z] == nodata_value:
                    result[idx_y, idx_x, idx_z] = nodata_value
                    continue

                for idx_h in range(0, hood_size):
                    hood_x = idx_x + offsets[idx_h][0]
                    hood_y = idx_y + offsets[idx_h][1]
                    hood_z = idx_z + offsets[idx_h][2]

                    if hood_x < 0 or hood_x >= arr.shape[1]:
                        if normalise_edges:
                            hood_normalise = True
                        continue

                    if hood_y < 0 or hood_y >= arr.shape[0]:
                        if normalise_edges:
                            hood_normalise = True
                        continue

                    if hood_z < 0 or hood_z >= arr.shape[2]:
                        if normalise_edges:
                            hood_normalise = True
                        continue

                    if nodata and arr[hood_y, hood_x, hood_z] == nodata_value:
                        if normalise_edges:
                            hood_normalise = True
                        continue

                    hood_values[hood_count] = arr[hood_y, hood_x, hood_z]
                    hood_weights[hood_count] = weights[idx_h]
                    hood_count += 1

                    if offsets[idx_h][0] == 0 and offsets[idx_h][1] == 0 and offsets[idx_h][2] == 0:
                        center_idx = hood_count - 1

                if hood_count == 0:
                    result[idx_y, idx_x, idx_z] = nodata_value
                    continue

                hood_values = hood_values[:hood_count]
                hood_weights = hood_weights[:hood_count]

                if hood_normalise:
                    hood_weights /= np.sum(hood_weights)

                result[idx_y, idx_x, idx_z] = hood_to_value(method, hood_values, hood_weights, nodata_value, center_idx, value)

    return result


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def convolve_array_simple2(
    array: np.ndarray,
    offsets: np.ndarray,
    weights: np.ndarray,
    nodata_value: float = -9999.9,
    intensity: float = 1.0,
):
    """
    Convolve a kernel with an array using a simple method.

    Args:
        array (np.ndarray): The array to convolve.
        offsets (np.ndarray): The offsets of the kernel.
        weights (np.ndarray): The weights of the kernel.
    
    Keyword Args:
        intensity (float=1.0): The intensity of the convolution. If
            1.0, the convolution is applied as is. If 0.5, the
            convolution is applied at half intensity.

    Returns:
        np.ndarray: The convolved array.
    """
    result = np.empty_like(array, dtype=np.float32)

    if intensity <= 0.0:
        return array.astype(np.float32)

    for col in prange(array.shape[0]):
        for row in prange(array.shape[1]):
            if array[col, row] == nodata_value:
                result[col, row] = nodata_value
                continue

            result_value = 0.0
            result_weight = 0.0
            for i in range(offsets.shape[0]):
                new_col = col + offsets[i, 0]
                new_row = row + offsets[i, 1]

                if new_col < 0 or new_col >= array.shape[0]:
                    continue

                if new_row < 0 or  new_row >= array.shape[1]:
                    continue

                result_weight += weights[i]
                result_value += array[new_col, new_row] * weights[i]

            if result_weight > 0.0:
                result_value /= result_weight
            else:
                result_value = nodata_value

    if intensity < 1.0:
        result *= intensity
        array *= (1.0 - intensity)
        result += array

    return result



@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def convolve_array_simple(
    array: np.ndarray,
    offsets: np.ndarray,
    weights: np.ndarray,
    intensity: float = 1.0,
):
    """
    Convolve a kernel with an array using a simple method.

    Args:
        array (np.ndarray): The array to convolve.
        offsets (np.ndarray): The offsets of the kernel.
        weights (np.ndarray): The weights of the kernel.
    
    Keyword Args:
        intensity (float=1.0): The intensity of the convolution. If
            1.0, the convolution is applied as is. If 0.5, the
            convolution is applied at half intensity.

    Returns:
        np.ndarray: The convolved array.
    """
    result = np.empty_like(array, dtype=np.float32)

    if intensity <= 0.0:
        return array.astype(np.float32)

    for col in prange(array.shape[0]):
        for row in prange(array.shape[1]):

            result_value = 0.0
            for i in range(offsets.shape[0]):
                new_col = col + offsets[i, 0]
                new_row = row + offsets[i, 1]

                if new_col < 0:
                    new_col = 0
                elif new_col >= array.shape[0]:
                    new_col = array.shape[0] - 1

                if new_row < 0:
                    new_row = 0
                elif new_row >= array.shape[1]:
                    new_row = array.shape[1] - 1

                result_value += array[new_col, new_row] * weights[i]

            result[col, row] = result_value

    if intensity < 1.0:
        result *= intensity
        array *= (1.0 - intensity)
        result += array

    return result



def convolve_array(
    arr: np.ndarray,
    offsets: np.ndarray,
    weights: np.ndarray,
    method: int = 1,
    nodata: bool = False,
    nodata_value: float = -9999.9,
    normalise_edges: bool = True,
    collapse: bool = False,
    value: Union[int, float, None] = None,
) -> np.ndarray:
    """
    Convolve an image with a function.

    Args:
        arr (numpy.ndarray): The input array to convolve.
        offsets (list of tuples): The list of offsets for the neighborhood
            used in the convolution.
        weights (list): The list of weights used in the convolution.

    Keyword Args:
        method (int=1): The method to use for the convolution.
            1: hood_sum
            2: hood_mode
            3: hood_max
            4: hood_min
            5: hood_contrast
            6: hood_quantile
            7: hood_standard_deviation
            8: hood_median_absolute_deviation
            9: hood_z_score
            10: hood_z_score_mad
            11: hood_sigma_lee
        nodata (bool=False): If True, nodata values are considered
            in the convolution.
        nodata_value (float=-9999.9): The value representing nodata.
        normalise_edges (bool=True): If True, the weights at the edges
            are normalised to sum to one. Only relavant for border pixels.
            Use false, if you are interested in the sum, otherwise you likely
            want to use True.
        collapse (bool=False): If True, the convolution results in a (height, width, 1)
            array. Otherwise, the convolution results in a (height, width, depth) applied
            channelwise.
        value (Union[int, float, None]=None): If not None, the value to use for the convolution.
            depending on the method specified.


    Returns:
        numpy.ndarray: The convolved array.
    """
    core_utils.type_check(arr, [np.ndarray], "arr")
    core_utils.type_check(offsets, [np.ndarray], "offsets")
    core_utils.type_check(weights, [np.ndarray], "weights")
    core_utils.type_check(method, [int], "method")
    core_utils.type_check(nodata, [bool], "nodata")
    core_utils.type_check(nodata_value, [float], "nodata_value")
    core_utils.type_check(normalise_edges, [bool], "normalise_edges")
    core_utils.type_check(collapse, [bool], "collapse")
    core_utils.type_check(value, [int, float, type(None)], "value")

    assert len(offsets) == len(weights), "offsets and weights must be the same length"
    assert method in range(1, len(METHOD_ENUMS)), "method must be between 1 and 11"
    assert arr.ndim in [2, 3], "arr must be 2 or 3 dimensional"

    if value is None:
        value = 0.5

    if arr.ndim == 2:
        arr = arr[:, :, np.newaxis]

    if collapse:
        return _convolve_array_collapse(
            arr,
            offsets,
            weights,
            method=method,
            nodata=nodata,
            nodata_value=nodata_value,
            normalise_edges=normalise_edges,
            value=value,
        )

    return _convolve_array(
        arr,
        offsets,
        weights,
        method=method,
        nodata=nodata,
        nodata_value=nodata_value,
        normalise_edges=normalise_edges,
        value=value,
    )



"""
This module contains function to select features from a dataset for machine learning.
"""

# Standard library
import sys; sys.path.append("../../")
from typing import Optional, Union, Tuple

# External
import numpy as np


def train_val_split(
    X: np.ndarray,
    y: np.ndarray,
    val_size: float = 0.2,
    random_state: Optional[Union[float, int]] = None,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Split a dataset into training and validation sets. Always shuffles.
    
    Args:
        X (np.ndarray): The input data.
        y (np.ndarray): The target data.
    
    Keyword Args:
        val_size (float=0.2): The size of the validation set as a fraction of the total dataset.
        random_state (float=None): The random seed to use for the split.

    Returns:
        x_train, x_val, y_train, y_val (np.ndarray): The training and validation sets in a tuple.
    """
    assert isinstance(X, np.ndarray), "X should be a numpy array."
    assert isinstance(y, np.ndarray), "y should be a numpy array."
    assert X.shape[0] == y.shape[0], "X and y should be the same axis=0 size."
    assert 0 < val_size < 1, "val_size should be between 0 and 1."

    if random_state is not None:
        np.random.seed(random_state)

    sample_indices = np.random.permutation(X.shape[0])

    split_idx = int(X.shape[0] * (1 - val_size))

    X_train = X[sample_indices[:split_idx]]
    X_test = X[sample_indices[split_idx:]]
    y_train = y[sample_indices[:split_idx]]
    y_test = y[sample_indices[split_idx:]]

    return X_train, X_test, y_train, y_test


def train_val_test_split(
    X: np.ndarray,
    y: np.ndarray,
    val_size: float = 0.1,
    test_size: float = 0.2,
    random_state: Optional[Union[float, int]] = None,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Split a dataset into training, validation, and test sets. Always shuffles.

    Args:
        X (np.ndarray): The input data.
        y (np.ndarray): The target data.

    Keyword Args:
        val_size (float=0.1): The size of the validation set as a fraction of the total dataset.
        test_size (float=0.2): The size of the test set as a fraction of the total dataset.
        random_state (float=None): The random seed to use for the split.
    
    Returns:
        x_train, x_val, x_test, y_train, y_val, y_test (np.ndarray): The training, validation, and test sets in a tuple.
    """
    assert isinstance(X, np.ndarray), "X should be a numpy array."
    assert isinstance(y, np.ndarray), "y should be a numpy array."
    assert X.shape[0] == y.shape[0], "X and y should be the same axis=0 size."
    assert 0 < val_size < 1, "val_size should be between 0 and 1."
    assert 0 < test_size < 1, "test_size should be between 0 and 1."

    if random_state is not None:
        np.random.seed(random_state)

    sample_indices = np.random.permutation(X.shape[0])

    test_split_idx = int(X.shape[0] * (1 - test_size))
    val_split_idx = int((1 - val_size) * test_split_idx)

    X_train = X[sample_indices[:val_split_idx]]
    X_val = X[sample_indices[val_split_idx:test_split_idx]]
    X_test = X[sample_indices[test_split_idx:]]
    y_train = y[sample_indices[:val_split_idx]]
    y_val = y[sample_indices[val_split_idx:test_split_idx]]
    y_test = y[sample_indices[test_split_idx:]]

    return X_train, X_val, X_test, y_train, y_val, y_test


def stratified_sampling(
    X: np.ndarray,
    y: np.ndarray,
    regression: bool = False,
    samples_per_class: bool = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Stratified sampling of a dataset. Can be used for regression or classification.
    
    Args:
        X (np.ndarray): The input data.
        y (np.ndarray): The target data.
    
    Keyword Args:
        regression (bool=False): Whether the dataset is for regression or classification.
        samples_per_class (int=None): The number of samples to take per class. If None,
            takes the same number of samples per class as the smallest class.

    Returns:
        X, y (np.ndarray): The stratified dataset.
    """
    assert isinstance(X, np.ndarray), "X should be a numpy array."
    assert isinstance(y, np.ndarray), "y should be a numpy array."
    assert X.shape[0] == y.shape[0], "X and y should be the same axis=0 size."
    assert samples_per_class is None or samples_per_class > 0, "samples_per_class should be greater than 0."

    y_classes = y

    if regression:
        assert y.dtype in [np.float64, np.float32, np.float16], "y should be a float array for regression."

        y_classes = np.digitize(y, np.percentile(y, [10, 20, 30, 40, 50, 60, 70, 80, 90]))

    unique_classes, counts = np.unique(y_classes, return_counts=True)

    if samples_per_class is None:
        samples_per_class = counts.min()

    stratified_indices = np.array([], dtype=int)

    for class_ in unique_classes:
        class_indices = np.where(y_classes == class_)[0]
        selected_indices = np.random.choice(class_indices, samples_per_class, replace=False)
        stratified_indices = np.concatenate((stratified_indices, selected_indices))

    return X[stratified_indices], y[stratified_indices]



"""
### Split functions for vector layers. ###

Dissolve vectors by attributes or geometry.
"""

# Standard library
import sys; sys.path.append("../../")
import os
from typing import Optional, Union
from uuid import uuid4

# External
from osgeo import ogr

# Internal
from buteo.utils import core_utils
from buteo.vector import core_vector


def split_vector_by_fid(
    vector: Union[str, ogr.DataSource],
    out_folder: Optional[str] = None,
    prefix: str = "",
    suffix: str = "",
) -> str:
    """
    Split a vector by feature id.

    Args:
        vector (ogr.DataSource): Input vector.

    Keyword Args:
        out_folder (str=None): Output folder.
        prefix (str=''): Prefix for output files.
        suffix (str=''): Suffix for output files.
    
    Returns:
        str: Output paths.
    """
    core_utils.type_check(vector, [ogr.DataSource, str], "vector")
    core_utils.type_check(out_folder, [str, None], "out_folder")
    assert core_utils.folder_exists(out_folder) or out_folder is None, "out_folder does not exist."

    opened = core_vector.open_vector(vector)
    metadata = core_vector.vector_to_metadata(opened)

    out_paths = []

    for layer_index in range(metadata["layer_count"]):
        layer = opened.GetLayerByIndex(layer_index)
        feature_count = layer.GetFeatureCount()

        layer.ResetReading()
        for _ in range(feature_count):
            feature = layer.GetNextFeature()

            if feature is None:
                continue

            feature_id = feature.GetFID()

            if out_folder is None:
                out_path = f"/vsimem/{prefix}{str(uuid4().int)}_{layer_index}_{feature_id}{suffix}.gpkg"
            else:
                out_path = os.path.join(
                    out_folder,
                    f"{prefix}{layer_index}_{feature_id}{suffix}.gpkg",
                )

            out_driver = ogr.GetDriverByName("GPKG")
            out_ds = out_driver.CreateDataSource(out_path)
            out_layer = out_ds.CreateLayer(
                layer.GetName(),
                srs=layer.GetSpatialRef(),
                geom_type=layer.GetGeomType(),
            )

            out_layer.CreateFeature(feature)
            out_layer.SyncToDisk()

            out_ds.FlushCache()
            out_ds, out_layer, feature = (None, None, None)

            out_paths.append(out_path)

    return out_paths



"""
This module contains functions for augmenting images that are
suited to remote sensing imagery.
"""
# Standard library
import sys; sys.path.append("../../")
from typing import Optional, Tuple, Any

# External
import numpy as np
from numba import jit, prange

# Internal
from buteo.ai.augmentation_funcs import (
    augmentation_rotation,
    augmentation_mirror,
    augmentation_noise,
    augmentation_channel_scale,
    augmentation_contrast,
    augmentation_drop_pixel,
    augmentation_drop_channel,
    augmentation_blur,
    augmentation_sharpen,
    augmentation_misalign,
    augmentation_cutmix,
    augmentation_mixup,
)


@jit(nopython=True, nogil=True, cache=True)
def _augmentation_batch_default(
    X: np.ndarray,
    y: Optional[np.ndarray],
    chance: float = 0.5,
    max_images: float = 1.0,
) -> Tuple[np.ndarray, Optional[np.ndarray], np.ndarray]:
    """ INTERNAL FUNCTION. DO NOT USE. """
    batch_size = X.shape[0]

    if chance == 0.0:
        aug_flags = np.zeros(batch_size, dtype=np.uint8)
        return X, y, aug_flags
    elif chance == 1.0:
        aug_flags = np.ones(batch_size, dtype=np.uint8)
        return X, y, aug_flags

    n_mixes = min(
        (np.random.rand(batch_size) <= chance).sum(),
        int(batch_size * max_images),
    )

    if n_mixes == 0:
        aug_flags = np.zeros(batch_size, dtype=np.uint8)
        return X, y, aug_flags

    aug_flags = np.zeros(batch_size, dtype=np.uint8)

    idx_targets = np.random.choice(batch_size, n_mixes, replace=False)
    for i in aug_flags:
        if i in idx_targets:
            aug_flags[i] = 1

    return X, y, aug_flags


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def augmentation_batch_rotation(
    X: np.ndarray,
    y: Optional[np.ndarray],
    chance: float = 0.5,
    max_images: float = 1.0,
    channel_last: bool = True,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Randomly rotates images in a batch by 90 degrees intervals. Images
    can be (batch, channels, height, width) or (batch, height, width, channels).

    Args:
        X (np.ndarray): The batch of images to rotate.
        y (np.ndarray/None): The label of images to rotate. If None, no label is returned.

    Keyword Args:
        y (np.ndarray/none=None): The batch of labels to rotate.
        chance (float=0.5): The chance of rotating the image.
        max_images (float=None): The maximum proportion of the images in the batch to possibly rotate.
        channel_last (bool=True): Whether the image is (channels, height, width) or (height, width, channels).

    Returns:
        Tuple[np.ndarray, Optional[np.ndarray]]: The rotated images and optionally labels.
    """
    x_rotated, y_rotated, selection = _augmentation_batch_default(X, y, chance, max_images)

    for idx in prange(selection.shape[0]):
        selected = selection[idx]

        if selected != np.uint8(1):
            continue

        if y is None:
            x_rotated[idx], _ = augmentation_rotation(
                X[idx], None,
                chance=1.0,
                k=None,
                channel_last=channel_last,
            )

        else:
            x_rotated[idx], y_rotated[idx] = augmentation_rotation(
                X[idx], y[idx],
                chance=1.0,
                k=None,
                channel_last=channel_last,
            )

    return x_rotated, y_rotated


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def augmentation_batch_mirror(
    X: np.ndarray,
    y: Optional[np.ndarray],
    chance: float = 0.5,
    max_images: float = 1.0,
    channel_last: bool = True,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Randomly mirrors images in a batch. Images can be (batch, channels, height, width) or (batch, height, width, channels).

    Args:
        X (np.ndarray): The batch of images to mirror.
        y (np.ndarray/None): The labels of images to mirror. If None, no label is returned.

    Keyword Args:
        chance (float=0.5): The chance of mirroring the image.
        max_images (float=1.0): The maximum proportion  of images in the batch to possibly mirror.
        k (int=None): If None, randomly mirrors the image along the horizontal or vertical axis.
            If 1, mirrors the image along the horizontal axis.
            If 2, mirrors the image along the vertical axis.
        channel_last (bool=True): Whether the image is (channels, height, width) or (height, width, channels).

    Returns:
        Tuple[np.ndarray, Optional[np.ndarray]]: The mirrored images and optionally labels.
    """
    x_mirrored, y_mirrored, selection = _augmentation_batch_default(X, y, chance, max_images)

    for idx in prange(selection.shape[0]):
        selected = selection[idx]

        if selected != np.uint8(1):
            continue

        if y is None:
            x_mirrored[idx], _ = augmentation_mirror(
                X[idx], None,
                chance=1.0,
                k=None,
                channel_last=channel_last,
            )

        else:
            x_mirrored[idx], y_mirrored[idx] = augmentation_mirror(
                X[idx], y[idx],
                chance=1.0,
                k=None,
                channel_last=channel_last,
            )

    return x_mirrored, y_mirrored


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def augmentation_batch_noise(
    X: np.ndarray,
    y: Optional[np.ndarray],
    chance: float = 0.5,
    max_amount: float = 0.1,
    max_images: float = 1.0,
    additive: bool = False,
    channel_last: Any = None, # pylint: disable=unused-argument
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Adds random noise seperately to each channel of the batch of images. The noise works
    for both channel first and last images.
    input should be (batch, height, width, channels) or (batch, channels, height, width).

    Args:
        X (np.ndarray): The batch of images to add noise to.
        y (np.ndarray/None): The labels of images to add noise to. If None, no label is returned.

    Keyword Args:
        chance (float=0.5): The chance of adding noise.
        amount (float=0.01): The amount of noise to add.
        max_amount (float=0.1): The maximum amount of noise to add, sampled uniformly.
        max_images (float=1.0): The maximum proportion of images in the batch to add noise to.
        additive (bool=False): Whether to add or multiply the noise.
        channel_last (any=None): Whether the image is (channels, height, width) or (height, width, channels).
            ignored for this function. Kept to keep the same function signature as other augmentations.

    Returns:
        Tuple[np.ndarray, Optional[np.ndarray]]: The batch of images with noise and optionally the unmodified label.
    """
    x_noised, y_noised, selection = _augmentation_batch_default(X, y, chance, max_images)

    for idx in prange(selection.shape[0]):
        selected = selection[idx]

        if selected != np.uint8(1):
            continue

        x_noised[idx], _ = augmentation_noise(
            X[idx], None,
            chance=1.0,
            max_amount=max_amount,
            additive=additive,
            channel_last=channel_last,
        )

    return x_noised, y_noised


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def augmentation_batch_channel_scale(
    X: np.ndarray,
    y: Optional[np.ndarray],
    chance: float = 0.5,
    max_amount: float = 0.1,
    max_images: Optional[float] = 1.0,
    additive: bool = False,
    channel_last: bool = True,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Scales the channels of the batch of images seperately by a fixed amount.
    input should be (batch, height, width, channels) or (batch, channels, height, width).

    Args:
        X (np.ndarray): The batch of images to scale the channels of.
        y (np.ndarray/None): The labels of images to scale the channels of. If None, no label is returned.

    Keyword Args:
        chance (float=0.5): The chance of scaling the channels.
        max_amount (float=0.1): The amount to possible scale the channels by. Sampled uniformly.
        max_images (float=1.0): The maximum number of images to add noise to.
        additive (bool=False): Whether to add or multiply the scaling.
        channel_last (bool=True): Whether the image is (channels, height, width) or (height, width, channels).

    Returns:
        Tuble[np.ndarray, Optional[np.ndarray]]: The batch of images with scaled channels and optionally the unmodified label.
    """
    x_scaled, y_scaled, selection = _augmentation_batch_default(X, y, chance, max_images)

    for idx in prange(selection.shape[0]):
        selected = selection[idx]

        if selected != np.uint8(1):
            continue

        x_scaled[idx], _ = augmentation_channel_scale(
            X[idx], None,
            chance=1.0,
            max_amount=max_amount,
            additive=additive,
            channel_last=channel_last,
        )

    return x_scaled, y_scaled


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def augmentation_batch_contrast(
    X: np.ndarray,
    y: Optional[np.ndarray],
    chance: float = 0.5,
    max_amount: float = 0.1,
    max_images: float = 1.0,
    channel_last: bool = True,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Changes the contrast of a batch of images by a random amount, seperately for each channel.
    input should be (batch, height, width, channels) or (batch, channels, height, width).

    Args:
        X (np.ndarray): The batch of images to change the contrast of.
        y (np.ndarray/None): The labels of images to change the contrast of. If None, no label is returned.

    Keyword Args:
        chance (float=0.5): The chance of changing the contrast.
        max_amount (float=0.01): The amount to change the contrast by.
        max_images (float=1.0): The maximum proportion of images in the batch to change the contrast of.
        channel_last (bool=True): Whether the image is (channels, height, width) or (height, width, channels).

    Returns:
        Tuble[np.ndarray, Optional[np.ndarray]]: The batch of images with changed contrast and optionally the unmodified label.
    """
    x_constrast, y_contrast, selection = _augmentation_batch_default(X, y, chance, max_images)

    for idx in prange(selection.shape[0]):
        selected = selection[idx]

        if selected != np.uint8(1):
            continue

        x_constrast[idx], _ = augmentation_contrast(
            X[idx], None,
            chance=1.0,
            max_amount=max_amount,
            channel_last=channel_last,
        )

    return x_constrast, y_contrast


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def augmentation_batch_drop_pixel(
    X: np.ndarray,
    y: Optional[np.ndarray],
    chance: float = 0.5,
    drop_probability: float = 0.01,
    drop_value: float = 0.0,
    max_images: float = 1.0,
    channel_last: bool = True,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Drops a random pixels from a batch of images.
    input should be (batch, height, width, channels) or (batch, channels, height, width).

    Args:
        X (np.ndarray): The batch of images to drop a pixel from.
        y (np.ndarray/None): The labels of images to drop pixels from. If None, no label is returned.

    Keyword Args:
        chance (float=0.5): The chance of dropping a pixel.
        drop_probability (float=0.05): The probability of dropping a pixel.
        drop_value (float=0.0): The value to drop the pixel to.
        max_images (float=1.0): The maximum proportion of images in the batch to drop a pixel from.
        channel_last (bool=True): Whether the image is (channels, height, width) or (height, width, channels).

    Returns:
        Tuple[np.ndarray, Optional[np.ndarray]]: The batch of images with the dropped pixels and optionally the unmodified label.
    """
    x_drop, y_drop, selection = _augmentation_batch_default(X, y, chance, max_images)

    for idx in prange(selection.shape[0]):
        selected = selection[idx]

        if selected != np.uint8(1):
            continue

        x_drop[idx], _ = augmentation_drop_pixel(
            X[idx], None,
            chance=1.0,
            drop_probability=drop_probability,
            drop_value=drop_value,
            channel_last=channel_last,
        )

    return x_drop, y_drop


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def augmentation_batch_drop_channel(
    X: np.ndarray,
    y: Optional[np.ndarray],
    chance: float = 0.5,
    drop_probability: float = 0.1,
    drop_value: float = 0.0,
    max_images: float = 1.0,
    channel_last: bool = True,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Drops a random channel from a batch of images.
    input should be (batch, height, width, channels) or (batch, channels, height, width).
    A maximum of one channel will be dropped.

    Args:
        X (np.ndarray): The batch of images to drop a channel from.
        y (np.ndarray/None): The labels of images to drop channels from. If None, no label is returned.

    Keyword Args:
        chance (float=0.5): The chance of dropping a channel.
        max_images (float=1.0): The maximum proportion of images in the batch to drop a channel from.
        drop_probability (float=0.1): The probability of dropping a channel.
        drop_value (float=0.0): The value to drop the channel to.
        channel_last (bool=True): Whether the image is (channels, height, width) or (height, width, channels).
    """
    x_drop, y_drop, selection = _augmentation_batch_default(X, y, chance, max_images)

    for idx in prange(selection.shape[0]):
        selected = selection[idx]

        if selected != np.uint8(1):
            continue

        x_drop[idx], _ = augmentation_drop_channel(
            X[idx], None,
            chance=1.0,
            drop_probability=drop_probability,
            drop_value=drop_value,
            channel_last=channel_last,
        )

    return x_drop, y_drop


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def augmentation_batch_blur(
    X: np.ndarray,
    y: Optional[np.ndarray],
    chance: float = 0.5,
    max_images: float = 1.0,
    intensity: float = 1.0,
    apply_to_y: bool = False,
    channel_last: bool = True,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Blurs a batch of images at random.
    input should be (batch, height, width, channels) or (batch, channels, height, width).

    Args:
        X (np.ndarray): The batch of images to potentially blur.
        y (np.ndarray/None): The labels of images to potentially blur. If None, no label is returned.

    Keyword Args:
        chance (float=0.5): The chance of blurring a pixel.
        max_images (float=1.0): The maximum proportion of images in the batch to blur.
        intensity (float=1.0): The intensity of the blur. from 0.0 to 1.0.
        apply_to_y (bool=False): Whether to blur the label as well.
        channel_last (bool=True): Whether the image is (channels, height, width) or (height, width, channels).
    """
    x_blurred, y_blurred, selection = _augmentation_batch_default(X, y, chance, max_images)

    for idx in prange(selection.shape[0]):
        selected = selection[idx]

        if selected != np.uint8(1):
            continue

        if apply_to_y and y is not None:
            y_blurred[idx], _  = augmentation_blur(
                y[idx], None,
                chance=1.0,
                intensity=intensity,
                apply_to_y=False,
                channel_last=channel_last,
            )

        x_blurred[idx], _ = augmentation_blur(
            X[idx], None,
            chance=1.0,
            intensity=intensity,
            apply_to_y=False,
            channel_last=channel_last,
        )

    return x_blurred, y_blurred


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def augmentation_batch_sharpen(
    X: np.ndarray,
    y: Optional[np.ndarray],
    chance: float = 0.5,
    max_images: float = 1.0,
    intensity: float = 1.0,
    apply_to_y: bool = False,
    channel_last: bool = True,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Sharpens a batch of images at random.
    input should be (batch, height, width, channels) or (batch, channels, height, width).

    Args:
        X (np.ndarray): The batch of images to potentially sharpen.
        y (np.ndarray/None): The labels of images to potentially sharpen. If None, no label is returned.

    Keyword Args:
        chance (float=0.5): The chance of sharpening a pixel.
        max_images (float=1.0): The maximum proportion of images in the batch to sharpen.
        intensity (float=1.0): The intensity of the sharpening. from 0.0 to 1.0.
        apply_to_y (bool=False): Whether to sharpen the label as well.
        channel_last (bool=True): Whether the image is (channels, height, width) or (height, width, channels).
    """
    x_sharpened, y_sharpened, selection = _augmentation_batch_default(X, y, chance, max_images)

    for idx in prange(selection.shape[0]):
        selected = selection[idx]

        if selected != np.uint8(1):
            continue

        if apply_to_y and y is not None:
            y_sharpened[idx], _ = augmentation_sharpen(
                y[idx], None,
                chance=1.0,
                intensity=intensity,
                apply_to_y=False,
                channel_last=channel_last,
            )

        x_sharpened[idx], _ = augmentation_sharpen(
            X[idx], None,
            chance=1.0,
            intensity=intensity,
            apply_to_y=False,
            channel_last=channel_last,
        )

    return x_sharpened, y_sharpened


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def augmentation_batch_misalign(
    X: np.ndarray,
    y: Optional[np.ndarray],
    chance: float = 0.5,
    max_offset: float = 0.5,
    max_images: float = 0.2,
    channel_last: bool = True,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Misaligns the channels of a batch of images at random.
    input should be (batch, height, width, channels) or (batch, channels, height, width).

    Args:
        X (np.ndarray): The batch of images to potentially misalign the channels of.
        y (np.ndarray/None): The labels of images to potentially misalign the channels of. If None, no label is returned.

    Keyword Args:
        chance (float=0.5): The chance of misaligning the channels of a pixel.
        max_offset (float=0.5): The maximum offset to misalign the channels by.
        max_images (float=0.2): The maximum number of images to misalign the channels of.
        max_channels (int=1): The maximum number of channels to misalign.
        channel_last (bool=True): Whether the image is (channels, height, width) or (height, width, channels).
    """
    x_misaligned, y_misaligned, selection = _augmentation_batch_default(X, y, chance, max_images)

    for idx in prange(selection.shape[0]):
        selected = selection[idx]

        if selected != np.uint8(1):
            continue

        x_misaligned[idx], _ = augmentation_misalign(
            X[idx], None,
            chance=1.0,
            max_offset=max_offset,
            channel_last=channel_last,
        )

    return x_misaligned, y_misaligned


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def augmentation_batch_cutmix(
    X: np.ndarray,
    y: np.ndarray,
    chance: float = 0.5,
    min_size: float = 0.333,
    max_size: float = 0.666,
    max_images: float = 0.2,
    label_mix: int = 0,
    feather: bool = True,
    feather_dist: int = 3,
    channel_last: bool = True,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Cutmixes a batch of images at random.
    input should be (batch, height, width, channels) or (batch, channels, height, width).

    Args:
        X (np.ndarray): The batch of images to potentially cutmix.
        y (np.ndarray/None): The labels of images to potentially cutmix. If None, no label is returned.

    Keyword Args:
        chance (float=0.5): The chance of cutmixing a pixel.
        min_size (float=0.333): The minimum size of the patch to cutmix. In percentage of the image width.
        max_size (float=0.666): The maximum size of the patch to cutmix. In percentage of the image width.
        max_images (float=0.2): The maximum percentage of images in a batch to mixup.
        label_mix (int=0): if
            0 - The labels will be mixed by the weights.\n
            1 - The target label will be used.\n
            2 - The source label will be used.\n
            3 - The max of the labels will be used.\n
            4 - The min of the labels will be used.\n
            5 - The max of the image with the highest weight will be used.\n
            6 - The min of the image with the highest weight will be used.\n
            7 - The sum of the labels will be used.\n
        feather (bool=True): Whether to feather the edges of the cutmix.
        feather_dist (int=3): The distance to feather the edges of the cutmix in pixels
        channel_last (bool=True): Whether the image is (channels, height, width) or (height, width, channels).
    """
    x_cutmix = X.astype(np.float32)
    y_cutmix = y.astype(np.float32)

    batch_size = x_cutmix.shape[0]

    n_mixes = min(
        (np.random.rand(batch_size) <= chance).sum(),
        int(batch_size * max_images),
    )

    if n_mixes == 0:
        return X, y

    idx_targets = np.random.choice(batch_size, n_mixes, replace=False)

    for idx_target in idx_targets:
        target_x = x_cutmix[idx_target]
        target_y = y_cutmix[idx_target]

        source_idxs = np.array([idx for idx in range(batch_size) if idx != idx_target])
        idx_source = np.random.choice(source_idxs, 1, replace=False)[0]

        source_x = x_cutmix[idx_source]
        source_y = y_cutmix[idx_source]

        cutmixed_x, cutmixed_y = augmentation_cutmix(
            target_x, target_y,
            source_x, source_y,
            min_size=min_size,
            max_size=max_size,
            label_mix=label_mix,
            feather=feather,
            feather_dist=feather_dist,
            channel_last=channel_last,
        )

        x_cutmix[idx_target] = cutmixed_x
        y_cutmix[idx_target] = cutmixed_y

    return x_cutmix, y_cutmix


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def augmentation_batch_mixup(
    X: np.ndarray,
    y: Optional[np.ndarray],
    chance: float = 0.5,
    min_size: float = 0.333,
    max_size: float = 0.666,
    label_mix: int = 0,
    max_images: float = 0.2,
    channel_last: bool = True,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Mixups a batch of images at random. This works by doing a linear intepolation between
    two images in the batch and then adding a random weight to each image.

    Mixup involves taking two images and blending them together by randomly interpolating
    their pixel values. More specifically, suppose we have two images x and x' with their
    corresponding labels y and y'. To generate a new training example, mixup takes a
    weighted sum of x and x', such that the resulting image x^* = λx + (1-λ)x',
    where λ is a randomly chosen interpolation coefficient. The label for the new image
    is also a weighted sum of y and y' based on the same interpolation coefficient.

    input should be (batch, height, width, channels) or (batch, channels, height, width).

    Args:
        X (np.ndarray): The batch of images to potentially mixup.
        y (np.ndarray/None): The batch of labels to potentially mixup.

    Keyword Args:
        chance (float=0.5): The chance of mixuping a pixel.
        min_size (float=0.333): The minimum size of the patch to cutmix. In percentage of the image width.
        max_size (float=0.666): The maximum size of the patch to cutmix. In percentage of the image width.
        max_images (float=0.2): The maximum percentage of images in a batch to mixup.
        label_mix (int=0): if
            0 - The labels will be mixed by the weights.\n
            1 - The target label will be used.\n
            2 - The source label will be used.\n
            3 - The max of the labels will be used.\n
            4 - The min of the labels will be used.\n
            5 - The max of the image with the highest weight will be used.\n
            6 - The min of the image with the highest weight will be used.\n
            7 - The sum of the labels will be used.\n
        channel_last (bool=True): Whether the image is (channels, height, width) or (height, width, channels).
    """
    x_mixup = X.astype(np.float32)
    y_mixup = y.astype(np.float32)

    batch_size = x_mixup.shape[0]

    n_mixes = min(
        (np.random.rand(batch_size) <= chance).sum(),
        int(batch_size * max_images),
    )

    if n_mixes == 0:
        return X, y

    idx_targets = np.random.choice(batch_size, n_mixes, replace=False)

    for idx_target in idx_targets:
        target_x = x_mixup[idx_target]
        target_y = y_mixup[idx_target]

        source_idxs = np.array([idx for idx in range(batch_size) if idx != idx_target])
        idx_source = np.random.choice(source_idxs, 1, replace=False)[0]

        source_x = x_mixup[idx_source]
        source_y = y_mixup[idx_source]

        mixupped_x, mixupped_y = augmentation_mixup(
            target_x, target_y,
            source_x, source_y,
            min_size=min_size,
            max_size=max_size,
            label_mix=label_mix,
            chance=chance,
            channel_last=channel_last,
        )

        x_mixup[idx_target] = mixupped_x
        y_mixup[idx_target] = mixupped_y

    return x_mixup, y_mixup



"""
This module contains functions for augmenting images that are
suited to remote sensing imagery.
"""
# Standard library
import sys; sys.path.append("../../")
from typing import Optional, List, Callable

import numpy as np
import buteo as beo

from buteo.ai.augmentation_funcs import (
    augmentation_rotation,
    augmentation_mirror,
    augmentation_channel_scale,
    augmentation_noise,
    augmentation_contrast,
    augmentation_blur,
    augmentation_blur_xy,
    augmentation_sharpen,
    augmentation_sharpen_xy,
    augmentation_drop_pixel,
    augmentation_drop_channel,
    augmentation_misalign,
    augmentation_cutmix,
    augmentation_mixup,
)


class AugmentationDataset():
    """
    A dataset that applies augmentations to the data.

    Args:
        X (np.ndarray): The data to augment.
        y (np.ndarray): The labels for the data.

    Keyword Args:
        augmentations (list): The augmentations to apply.
        callback (callable): A callback to apply to the data after augmentation.
        input_is_channel_last (bool=True): Whether the data is in channel last format.
        output_is_channel_last (bool=False): Whether the output should be in channel last format.
    
    Returns:
        A dataset yielding batches of augmented data. For Pytorch,
            convert the batches to tensors before ingestion.
    """
    def __init__(self,
        X: np.ndarray,
        y: np.ndarray,
        augmentations: Optional[List] = None,
        callback: Callable = None,
        input_is_channel_last: bool = True,
        output_is_channel_last: bool = False,
    ):
        if input_is_channel_last:
            self.x_train = beo.channel_last_to_first(X)
            self.y_train = beo.channel_last_to_first(y)
        else:
            self.x_train = X
            self.y_train = y

        if augmentations is None:
            self.augmentations = []
        else:
            self.augmentations = augmentations

        self.callback = callback
        self.channel_last = False
        self.output_is_channel_last = output_is_channel_last

    def __len__(self):
        return len(self.x_train)

    def __getitem__(self, index):
        x = self.x_train[index]
        y = self.y_train[index]

        for aug in self.augmentations:
            aug_name = aug["name"]
            if aug_name == "rotation":
                func = augmentation_rotation
            elif aug_name == "mirror":
                func = augmentation_mirror
            elif aug_name == "channel_scale":
                func = augmentation_channel_scale
            elif aug_name == "noise":
                func = augmentation_noise
            elif aug_name == "contrast":
                func = augmentation_contrast
            elif aug_name == "drop_pixel":
                func = augmentation_drop_pixel
            elif aug_name == "drop_channel":
                func = augmentation_drop_channel
            elif aug_name == "blur":
                func = augmentation_blur
            elif aug_name == "blur_xy":
                func = augmentation_blur_xy
            elif aug_name == "sharpen":
                func = augmentation_sharpen
            elif aug_name == "sharpen_xy":
                func = augmentation_sharpen_xy
            elif aug_name == "misalign":
                func = augmentation_misalign
            elif aug_name == "cutmix":
                func = augmentation_cutmix
            elif aug_name == "mixup":
                func = augmentation_mixup

            if func is None:
                raise ValueError(f"Augmentation {aug['name']} not supported.")

            channel_last = self.channel_last
            kwargs = {key: value for key, value in aug.items() if key != "name"}

            if aug_name in ["cutmix", "mixup"]:
                idx_source = np.random.randint(len(self.x_train))
                xx = self.x_train[idx_source]
                yy = self.y_train[idx_source]

                x, y = func(x, y, xx, yy, channel_last=channel_last, **kwargs)
            else:
                x, y = func(x, y, channel_last=channel_last, **kwargs)

        if self.callback is not None:
            x, y = self.callback(x, y)

        if self.output_is_channel_last:
            x = beo.channel_first_to_last(x)
            y = beo.channel_first_to_last(y)

        return x, y



"""
Tools for working with sentinel data
"""
from .augmentation import *
from .augmentation_batch import *
from .augmentation_funcs import *
from .augmentation_utils import *
from .scalers import *
from .selection import *



"""
### Perform convolutions on arrays. (Funcs) ###
"""

# Standard Library
from typing import Union

# External
import numpy as np
from numba import jit


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def hood_max(
    values: np.ndarray,
    weights: np.ndarray,
) -> Union[float, int]:
    """ Get the weighted maximum. """
    idx = np.argmax(np.multiply(values, weights))
    return values[idx]


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def hood_min(
    values: np.ndarray,
    weights: np.ndarray,
) -> Union[float, int]:
    """ Get the weighted minimum. """
    max_val = values.max()
    adjusted_values = np.where(weights == 0.0, max_val, values)
    idx = np.argmin(np.divide(adjusted_values, weights + 1e-7))
    return values[idx]


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def hood_sum(
    values: np.ndarray,
    weights: np.ndarray,
) -> Union[float, int]:
    """ Get the weighted sum. """
    return np.sum(np.multiply(values, weights))


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def hood_mode(
    values: np.ndarray,
    weights: np.ndarray,
) -> Union[float, int]:
    """ Get the weighted sum. """
    values_ints = np.rint(values)
    unique = np.unique(values_ints)

    most_occured_value = 0
    most_occured_weight = -9999.9

    for unique_value in unique:
        cum_weight = 0
        for idx in range(values.shape[0]):
            if values_ints[idx] == unique_value:
                cum_weight += weights[idx]

        if cum_weight > most_occured_weight:
            most_occured_weight = cum_weight
            most_occured_value = unique_value

    return most_occured_value


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def hood_count_occurances(
    values: np.ndarray,
    weights: np.ndarray,
    value: Union[int, float, None],
    normalise: bool=False,
) -> Union[float, int]:
    """ Count how many times a number appears in an array. 
        Can be normalised to the size of the array to do feathering.
    """
    if value is None:
        return 0.0

    occurances = 0.0
    for idx in range(values.shape[0]):
        if values[idx] == value and weights[idx] > 0.0:
            occurances += weights[idx]

    if normalise:
        occurances = occurances / values.size

    return occurances


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def hood_contrast(
    values: np.ndarray,
    weights: np.ndarray,
) -> Union[float, int]:
    """ Get the local contrast. """
    max_val = values.max()
    adjusted_values = np.where(weights == 0.0, max_val, values)
    local_min = np.min(np.divide(adjusted_values, weights + 1e-7))
    local_max = np.max(np.multiply(values, weights))

    return np.abs(local_max - local_min)

@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def hood_quantile(
    values: np.ndarray,
    weights: np.ndarray,
    quantile: float,
) -> Union[float, int]:
    """ Get the weighted median. """
    sort_mask = np.argsort(values)
    sorted_data = values[sort_mask]
    sorted_weights = weights[sort_mask]
    cumsum = np.cumsum(sorted_weights)
    intersect = (cumsum - 0.5 * sorted_weights) / cumsum[-1]
    return np.interp(quantile, intersect, sorted_data)


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def hood_median_absolute_deviation(
    values: np.ndarray,
    weights: np.ndarray,
) -> Union[float, int]:
    """ Get the median absolute deviation """
    median = hood_quantile(values, weights, 0.5)
    absdeviation = np.abs(np.subtract(values, median))
    return hood_quantile(absdeviation, weights, 0.5)


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def hood_z_score(
    values: np.ndarray,
    weights: np.ndarray,
    center_idx: int,
) -> Union[float, int]:
    """ Get the local z score ."""
    std = hood_standard_deviation(values, weights)
    mean = hood_sum(values, weights)

    center_value = values[center_idx]

    return (center_value - mean) / std


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def hood_z_score_mad(
    values: np.ndarray,
    weights: np.ndarray,
    center_idx: int,
) -> Union[float, int]:
    """ Get the local z score calculated around the MAD. """
    mad_std = hood_median_absolute_deviation(values, weights) * 1.4826
    median = hood_quantile(values, weights, 0.5)

    center_value = values[center_idx]

    return (center_value - median) / mad_std


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def hood_standard_deviation(
    values: np.ndarray,
    weights: np.ndarray,
) -> Union[float, int]:
    "Get the weighted standard deviation. "
    summed = hood_sum(values, weights)
    variance = np.sum(np.multiply(np.power(np.subtract(values, summed), 2), weights))
    return np.sqrt(variance)


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def k_to_size(size: int) -> int:
    """ Preprocess Sigma Lee limits. """
    return int(np.rint(-0.0000837834 * size ** 2 + 0.045469 * size + 0.805733))


@jit(nopython=True, parallel=True, nogil=True, fastmath=True, cache=True)
def hood_sigma_lee(
    values: np.ndarray,
    weights: np.ndarray,
) -> Union[float, int]:
    """ Sigma lee SAR filter. """
    std = hood_standard_deviation(values, weights)
    selected_values = np.zeros_like(values)
    selected_weights = np.zeros_like(weights)

    sigma_mult = 1
    passed = 0
    attempts = 0
    ks = k_to_size(values.size)

    while passed < ks and attempts < 5:
        for idx, val in np.ndenumerate(values):
            if val >= std * sigma_mult or val <= -std * sigma_mult:
                selected_values[idx] = val
                selected_weights[idx] = weights[idx]
                passed += 1

        sigma_mult += 1
        attempts += 1

    if passed < ks:
        return hood_sum(values, weights)

    sum_of_weights = np.sum(selected_weights)

    if sum_of_weights == 0:
        return 0

    selected_weights = np.divide(selected_weights, sum_of_weights)

    return hood_sum(selected_values, selected_weights)


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def hood_roughness(
    values: np.ndarray,
    weights: np.ndarray,
    center_idx: int,
) -> Union[float, int]:
    """ 
        Defined as the maximum difference between the center value and the
        surrounding values. Weighted.
    """
    center_value = values[center_idx]

    max_idx = np.argmax(np.abs(np.subtract(values, center_value)) * weights)

    return np.abs(center_value - values[max_idx])


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def hood_roughness_tpi(
    values: np.ndarray,
    weights: np.ndarray,
    center_idx: int,
) -> Union[float, int]:
    """ 
        Defined as the difference between the center pixel and the mean of
        the surrounding pixels. Weighted.
    """
    center_value = values[center_idx]
    values_non_center = np.delete(values, center_idx)
    weights_non_center = np.delete(weights, center_idx)
    weights_non_center = np.divide(weights_non_center, np.sum(weights_non_center))

    return np.abs(center_value - hood_sum(values_non_center, weights_non_center))


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def hood_roughness_tri(
    values: np.ndarray,
    weights: np.ndarray,
    center_idx: int,
) -> Union[float, int]:
    """ 
        Defined as the mean difference between the center pixel and the
        surrounding pixels. Weighted.
    """
    center_value = values[center_idx]
    values_non_center = np.delete(values, center_idx)
    weights_non_center = np.delete(weights, center_idx)
    weights_non_center = np.divide(weights_non_center, np.sum(weights_non_center))

    return hood_sum(np.abs(np.subtract(values_non_center, center_value)), weights_non_center)



"""
This module contains utility functions for augmenting images that are
suited to remote sensing imagery.
"""
# Standard library
from typing import Tuple, List, Union

# External
import numpy as np
from numba import jit, prange

@jit(nopython=True, nogil=True, cache=True)
def is_integer_dtype(dtype: np.dtype) -> bool:
    """
    Check if a dtype is an integer type 
    
    Args:
        dtype (np.dtype): The dtype to check.

    Returns:
        bool: True if the dtype is an integer type.
    """
    return dtype.kind in ('i', 'u')


# These should have .astype(dtype, copy=False) to avoid copying the data
# But this is currently not supported by numba.
@jit(nopython=True, nogil=True, cache=True)
def fit_data_to_dtype(
    data: np.ndarray,
    dtype: np.dtype,
) -> np.ndarray:
    """
    Fit data to a dtype. If the dtype is an integer type, the data will be
        rounded to the nearest integer. If the dtype is a float type, the data
        will be clipped to the min and max values of the dtype.
    
    Args:
        data (np.ndarray): The data to fit to the dtype.
        dtype (np.dtype): The dtype to fit the data to.

    Returns:
        np.ndarray: The data fitted to the dtype.
    """
    # TODO: This is a workaround for numba. Remove when numba supports.
    return data.astype(dtype)

    # return data.astype(dtype)
    # max_value = np.iinfo(dtype).max
    # min_value = np.iinfo(dtype).min

    # if is_integer_dtype(dtype):
    #     return np.rint(
    #         np.clip(data, min_value, max_value)
    #     ).astype(dtype)

    # return np.clip(
    #     data,
    #     min_value,
    #     max_value,
    # ).astype(dtype)


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def feather_box_2d(
    array: np.ndarray,
    bbox: Union[List[int], Tuple[int]],
    feather_dist: int = 3,
) -> np.ndarray:
    """
    Feather a box into an array (2D). Box should be the original box
        buffered by feather_dist.

    Args:
        array_whole (np.ndarray): The array containing the box.
        bbox (Union[List[int], Tuple[int]]): The box.
            the bbox should be in the form [x_min, x_max, y_min, y_max].
    
    Keyword Args:
        feather_dist (int=3): The distance to feather the box.

    Returns:
        Tuple[np.ndarray, np.ndarray]: The featherweights for the array and the bbox.
    """
    # Get the bbox
    x_min, x_max, y_min, y_max = bbox

    kernel_size = int((feather_dist * 2) + 1)
    n_offsets = kernel_size ** 2
    max_dist = (np.sqrt(2) / 2) + feather_dist

    feather_offsets = np.zeros((n_offsets, 2), dtype=np.int64)

    within_circle = 0
    for i in prange(-feather_dist, feather_dist + 1):
        for j in prange(-feather_dist, feather_dist + 1):
            dist = np.sqrt(i ** 2 + j ** 2)
            if dist <= max_dist:
                feather_offsets[within_circle][0] = i
                feather_offsets[within_circle][1] = j
                within_circle += 1

    feather_offsets = feather_offsets[:within_circle]
    n_offsets = feather_offsets.shape[0]

    x_min = max(0, x_min - feather_dist)
    x_max = min(x_max + 1 + feather_dist, array.shape[1])
    y_min = max(0, y_min - feather_dist)
    y_max = min(y_max + 1 + feather_dist, array.shape[0])

    feather_weights_array = np.ones_like(array, dtype=np.float32)

    # Feather the box
    for y in prange(y_min, y_max + 1):
        for x in prange(x_min, x_max + 1):
            total_possible = n_offsets
            within_bbox = 0
            for offset in feather_offsets:
                x_offset, y_offset = offset
                x_new = x + x_offset
                y_new = y + y_offset

                if y_new < 0 or y_new >= array.shape[0]:
                    total_possible -= 1
                    continue

                if x_new < 0 or x_new >= array.shape[1]:
                    total_possible -= 1
                    continue

                if x_new - feather_dist >= x_min and x_new + feather_dist <= x_max and y_new - feather_dist >= y_min and y_new + feather_dist <= y_max:
                    within_bbox += 1

            weights_box = within_bbox / total_possible
            feather_weights_array[y, x] = 1 - weights_box

    return feather_weights_array


@jit(nopython=True, nogil=True, cache=True, fastmath=True, inline='always')
def rotate_90(
    arr: np.ndarray,
    channel_last: bool = True
) -> np.ndarray:
    """ Rotate a 3D array 90 degrees clockwise.
    
    Args:
        arr (np.ndarray): The array to rotate.
    
    Keyword Args:
        channel_last (bool=True): Whether the last axis is the channel axis.

    Returns:
        np.ndarray: The rotated array.
    """
    if channel_last:
        return arr[::-1, :, :].transpose(1, 0, 2) # (H, W, C)

    return arr[:, ::-1, :].transpose(0, 2, 1) # (C, H, W)


@jit(nopython=True, nogil=True, cache=True, fastmath=True, inline='always')
def rotate_180(
    arr: np.ndarray,
    channel_last: bool = True
) -> np.ndarray:
    """ Rotate a 3D array 180 degrees clockwise.

    Args:
        arr (np.ndarray): The array to rotate.
    
    Keyword Args:
        channel_last (bool=True): Whether the last axis is the channel axis.

    Returns:
        np.ndarray: The rotated array.
    """
    if channel_last:
        return arr[::-1, ::-1, :]  # (H, W, C)

    return arr[:, ::-1, ::-1] # (C, H, W)


@jit(nopython=True, nogil=True, cache=True, fastmath=True, inline='always')
def rotate_270(
    arr: np.ndarray,
    channel_last: bool = True
) -> np.ndarray:
    """ 
    Rotate a 3D image array 270 degrees clockwise.

    Args:
        arr (np.ndarray): The array to rotate.

    Keyword Args:
        channel_last (bool=True): Whether the last axis is the channel axis.

    Returns:
        np.ndarray: The rotated array.
    """
    if channel_last:
        return arr[:, ::-1, :].transpose(1, 0, 2) # (H, W, C)

    return arr[:, :, ::-1].transpose(0, 2, 1) # (C, H, W)


@jit(nopython=True, nogil=True, cache=True, fastmath=True, inline='always')
def rotate_arr(
    arr: np.ndarray,
    k: int,
    channel_last: bool = True,
) -> np.ndarray:
    """ Rotate an array by 90 degrees intervals clockwise.
    
    Args:
        arr (np.ndarray): The array to rotate.
        k (int): The number of 90 degree intervals to rotate by.
            1 for 90 degrees, 2 for 180 degrees, 3 for 270 degrees. 0 for no rotation.

    Keyword Args:
        channel_last (bool=True): Whether the last axis is the channel axis.

    Returns:
        np.ndarray: The rotated array.
    """
    if k == 0:
        return arr.copy()

    view = arr

    if k == 1:
        view = rotate_90(arr, channel_last=channel_last)
    elif k == 2:
        view = rotate_180(arr, channel_last=channel_last)
    elif k == 3:
        view = rotate_270(arr, channel_last=channel_last)

    return view.copy()


@jit(nopython=True, nogil=True, cache=True, fastmath=True, inline='always')
def mirror_horizontal(
    arr: np.ndarray,
    channel_last: bool = True
) -> np.ndarray:
    """ Mirror a 3D array horizontally.
    
    Args:
        arr (np.ndarray): The array to mirror.
    
    Keyword Args:
        channel_last (bool=True): Whether the last axis is the channel axis.

    Returns:
        np.ndarray: The mirrored array.
    """
    if channel_last:
        return arr[:, ::-1, :] # (H, W, C)

    return arr[:, :, ::-1] # (C, H, W)


@jit(nopython=True, nogil=True, cache=True, fastmath=True, inline='always')
def mirror_vertical(
    arr: np.ndarray,
    channel_last: bool = True
) -> np.ndarray:
    """ Mirror a 3D array vertically.

    Args:
        arr (np.ndarray): The array to mirror.

    Keyword Args:
        channel_last (bool=True): Whether the last axis is the channel axis.
    
    Returns:
        np.ndarray: The mirrored array.
    """
    if channel_last:
        return arr[::-1, :, :] # (H, W, C)

    return arr[:, ::-1, :] # (C, H, W)


@jit(nopython=True, nogil=True, cache=True, fastmath=True, inline='always')
def mirror_horisontal_vertical(
    arr: np.ndarray,
    channel_last: bool = True
) -> np.ndarray:
    """ Mirror a 3D array horizontally and vertically.

    Args:
        arr (np.ndarray): The array to mirror.

    Keyword Args:
        channel_last (bool=True): Whether the last axis is the channel axis.

    Returns:
        np.ndarray: The mirrored array.
    """
    if channel_last:
        return arr[::-1, ::-1, :] # (H, W, C)

    return arr[:, ::-1, ::-1] # (C, H, W)


@jit(nopython=True, nogil=True, cache=True, fastmath=True, inline='always')
def mirror_arr(
    arr: np.ndarray,
    k: int,
    channel_last: bool = True,
) -> np.ndarray:
    """ Mirror an array horizontally and/or vertically.

    Args:
        arr (np.ndarray): The array to mirror.
        k (int): 1 for horizontal, 2 for vertical, 3 for both, 0 for no mirroring.
    
    Keyword Args:
        channel_last (bool=True): Whether the last axis is the channel axis.
    """
    if k == 0:
        return arr.copy()

    view = arr

    if k == 1:
        view = mirror_horizontal(arr, channel_last=channel_last)
    elif k == 2:
        view = mirror_vertical(arr, channel_last=channel_last)
    elif k == 3:
        view = mirror_horisontal_vertical(arr, channel_last=channel_last)

    return view.copy()



"""
### Calculate distances on a raster. ###

Module to calculate the distance from a pixel value to other pixels.
"""

# Standard library
import sys; sys.path.append("../../")
from uuid import uuid4
from typing import Union, List

# External
import numpy as np
from osgeo import gdal

# Internal
from buteo.utils import core_utils, gdal_utils
from buteo.raster import core_raster
from buteo.raster.borders import add_border_to_raster



def calc_proximity(
    input_rasters: List,
    target_value: Union[int, float] = 1,
    unit: str = "GEO",
    out_path: Union[str, None, List] = None,
    max_dist: Union[int, float] = 1000,
    add_border: bool = False,
    border_value: Union[int, float] = 0,
    weighted: bool = False,
    invert: bool = False,
    return_array: bool = False,
    prefix: str = "",
    suffix: str = "_proximity",
    add_uuid: bool = False,
    overwrite: bool = True,
) -> Union[str, np.ndarray]:
    """
    Calculate the proximity of input_raster to values.

    Args:
        input_rasters (List): A list of rasters to use as input.
    
    Keyword Args:
        target_value (Union[int, float]=1): The value to use as target.
        unit (str='GEO'): The unit to use for the distance, GEO or PIXEL.
        out_path (Union[str, None, List]=None): The output path.
        max_dist (Union[int, float]=1000): The maximum distance to use.
        add_border (bool=False): If True, a border will be added to the raster.
        border_value (Union[int, float]=0): The value to use for the border.
        weighted (bool=False): If True, the distance will be divided by the max distance.
        invert (bool=False): If True, the target will be inversed.
        return_array (bool=False): If True, a NumPy array will be returned instead of a raster.
        prefix (str=''): Prefix to add to the output.
        suffix (str='_proximity'): Suffix to add to the output.
        add_uuid (bool=False): Should a uuid be added to the output path?
        overwrite (bool=True): If the output path exists already, should it be overwritten?

    Returns:
        Union[str, np.ndarray]: A path to a raster with the calculated proximity, or a numpy array with the data.
    """
    core_utils.type_check(input_rasters, [str, gdal.Dataset, [str, gdal.Dataset]], "input_rasters")
    core_utils.type_check(target_value, [int, float], "target_value")
    core_utils.type_check(out_path, [str, [str], None], "out_path")
    core_utils.type_check(max_dist, [int, float], "max_dist")
    core_utils.type_check(add_border, [bool], "add_border")
    core_utils.type_check(border_value, [int, float], "border_value")
    core_utils.type_check(weighted, [bool], "weighted")
    core_utils.type_check(invert, [bool], "invert")
    core_utils.type_check(return_array, [bool], "return_array")
    core_utils.type_check(prefix, [str], "prefix")
    core_utils.type_check(suffix, [str], "suffix")
    core_utils.type_check(add_uuid, [bool], "add_uuid")
    core_utils.type_check(overwrite, [bool], "overwrite")

    raster_list = core_utils.ensure_list(input_rasters)
    path_list = gdal_utils.create_output_path_list(
        raster_list,
        out_path=out_path,
        overwrite=overwrite,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
    )

    output = []
    for index, input_raster in enumerate(raster_list):
        out_path = path_list[index]

        in_arr = core_raster.raster_to_array(input_raster, filled=True)
        bin_arr = (in_arr != target_value).astype("uint8")
        bin_raster = core_raster.array_to_raster(bin_arr, reference=input_raster)

        in_raster = core_raster._open_raster(bin_raster)
        in_raster_path = bin_raster

        if add_border:
            border_size = 1
            border_raster = add_border_to_raster(
                in_raster,
                border_size=border_size,
                border_value=border_value,
                overwrite=True,
            )

            in_raster = core_raster._open_raster(border_raster)

            gdal.Unlink(in_raster_path)
            in_raster_path = border_raster

        src_band = in_raster.GetRasterBand(1)

        driver_name = "GTiff" if out_path is None else gdal_utils.path_to_driver_raster(out_path)
        if driver_name is None:
            raise ValueError(f"Unable to parse filetype from path: {out_path}")

        driver = gdal.GetDriverByName(driver_name)
        if driver is None:
            raise ValueError(f"Error while creating driver from extension: {out_path}")

        mem_path = f"/vsimem/raster_proximity_tmp_{uuid4().int}.tif"

        dest_raster = driver.Create(
            mem_path,
            in_raster.RasterXSize,
            in_raster.RasterYSize,
            1,
            gdal.GetDataTypeByName("Float32"),
        )

        dest_raster.SetGeoTransform(in_raster.GetGeoTransform())
        dest_raster.SetProjection(in_raster.GetProjectionRef())
        dst_band = dest_raster.GetRasterBand(1)

        gdal.ComputeProximity(
            src_band,
            dst_band,
            options=[
                "VALUES='1'",
                f"DISTUNITS={unit}",
                f"MAXDIST={max_dist}",
            ],
        )

        dst_arr = dst_band.ReadAsArray()
        gdal.Unlink(mem_path)
        gdal.Unlink(in_raster_path)

        dst_arr = np.where(dst_arr > max_dist, max_dist, dst_arr)

        if invert:
            dst_arr = max_dist - dst_arr

        if weighted:
            dst_arr = dst_arr / max_dist

        if add_border:
            dst_arr = dst_arr[border_size:-border_size, border_size:-border_size]

        src_band = None
        dst_band = None
        in_raster = None
        dest_raster = None

        if return_array:
            output.append(dst_arr)
        else:
            core_raster.array_to_raster(dst_arr, reference=input_raster, out_path=out_path)
            output.append(out_path)

        dst_arr = None

    if isinstance(input_rasters, list):
        return output

    return output[0]



"""
This module contains functions for augmenting images that are
suited to remote sensing imagery.
"""
# Standard library
import sys; sys.path.append("../../")
from typing import Optional, Tuple, Any

# External
import numpy as np
from numba import jit, prange

# Internal
from buteo.ai.augmentation_utils import (
    fit_data_to_dtype,
    feather_box_2d,
    rotate_arr,
    mirror_arr,
)
from buteo.raster.convolution import (
    convolve_array_simple,
    simple_blur_kernel_2d_3x3,
    simple_unsharp_kernel_2d_3x3,
    simple_shift_kernel_2d,
)


@jit(nopython=True, nogil=True, cache=True, fastmath=True)
def augmentation_rotation(
    X: np.ndarray,
    y: Optional[np.ndarray],
    chance: float = 0.2,
    k: Optional[int] = None,
    channel_last: bool = True,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Randomly rotate the image by 90 degrees intervals. Images
    can be (channels, height, width) or (height, width, channels).

    Args:
        X (np.ndarray): The image to rotate.
        y (np.ndarray/None): The label to rotate.

    Keyword Args:
        chance (float=0.2): The chance of rotating the image.
        k (int=None): The number of 90 degree intervals to rotate by.
        channel_last (bool=True): Whether the image is (channels, height, width) or (height, width, channels).

    Returns:
        Tuple[np.ndarray, Optional[np.ndarray]]: The rotated image and optionally the label.
    """
    if np.random.rand() > chance or k == 0:
        return X, y

    random_k = np.random.randint(1, 4) if k is None else k
    X_rot = rotate_arr(X, random_k, channel_last)

    if y is None:
        return X_rot, y

    y_rot = rotate_arr(y, random_k, channel_last)

    return X_rot, y_rot


@jit(nopython=True, nogil=True, cache=True, fastmath=True)
def augmentation_mirror(
    X: np.ndarray,
    y: Optional[np.ndarray],
    chance: float = 0.2,
    k: Optional[int] = None,
    channel_last: bool = True,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Randomly mirrors the image. Images can be (channels, height, width) or (height, width, channels).

    Args:
        X (np.ndarray): The image to mirror.
        y (np.ndarray/None): The label to mirror.

    Keyword Args:
        chance (float=0.2): The chance of mirroring the image.
        k (int=None): If None, randomly mirrors the image along the horizontal or vertical axis.
            If 1, mirrors the image along the horizontal axis.
            If 2, mirrors the image along the vertical axis.
            If 3, mirrors the image along both the horizontal and vertical axis.
        channel_last (bool=True): Whether the image is (channels, height, width) or (height, width, channels).

    Returns:
        Tuple[np.ndarray, Optional[np.ndarray]]: The mirrored image and optionally the label.
    """
    if np.random.rand() > chance or k == 0:
        return X, y

    random_k = np.random.randint(1, 4) if k is None else k

    flipped_x = mirror_arr(X, random_k, channel_last)
    flipped_y = mirror_arr(y, random_k, channel_last) if y is not None else None

    return flipped_x, flipped_y


@jit(nopython=True, nogil=True, cache=True, fastmath=True)
def augmentation_noise(
    X: np.ndarray,
    y: Optional[np.ndarray],
    chance: float = 0.2,
    max_amount: float = 0.1,
    additive: bool = False,
    channel_last: Any = None, # pylint: disable=unused-argument
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Adds random noise seperately to each pixel of the image. The noise works
    for both channel first and last images.
    input should be (height, width, channels) or (channels, height, width).

    Args:
        X (np.ndarray): The image to add noise to.
        y (np.ndarray/None): The label to add noise to. If None, no label is returned.

    Keyword Args:
        chance (float=0.2): The chance of adding noise.
        max_amount (float=0.1): The maximum amount of noise to add, sampled uniformly.
        additive (bool=False): Whether to add or multiply the noise.
        channel_last (bool=True): Whether the image is (channels, height, width) or (height, width, channels).
            ignored for this function. Kept to keep the same function signature as other augmentations.

    Returns:
        Tuple[np.ndarray, Optional[np.ndarray]]: The image with noise and optionally the unmodified label.
    """
    if np.random.rand() > chance:
        return X, y

    amount = np.random.rand() * max_amount

    if additive:
        noisy_x = X + np.random.normal(0, amount, X.shape)
    else:
        noisy_x = X * np.random.normal(1, amount, X.shape)

    return fit_data_to_dtype(noisy_x, X.dtype), y


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def augmentation_channel_scale(
    X: np.ndarray,
    y: Optional[np.ndarray],
    chance: float = 0.2,
    max_amount: float = 0.1,
    additive: bool = False,
    channel_last: bool = True,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Scales the channels of the image seperately by a fixed amount.
    input should be (height, width, channels) or (channels, height, width).

    Args:
        X (np.ndarray): The image to scale the channels of.
        y (np.ndarray/None): The label to scale the channels of. If None, no label is returned.

    Keyword Args:
        chance (float=0.2): The chance of scaling the channels.
        max_amount (float=0.1): The amount to possible scale the channels by. Sampled uniformly.
        additive (bool=False): Whether to add or multiply the scaling.
        channel_last (bool=True): Whether the image is (channels, height, width) or (height, width, channels).

    Returns:
        Tuple[np.ndarray, Optional[np.ndarray]]: The image with scaled channels and optionally the unmodified label.
    """
    if np.random.rand() > chance:
        return X, y

    x_scaled = X.astype(np.float32)
    y_scaled = y

    amount = np.random.rand() * max_amount

    if channel_last:
        for i in prange(X.shape[2]):
            if additive:
                random_amount = np.random.uniform(-amount, amount)
                x_scaled[:, :, i] += random_amount
            else:
                random_amount = np.random.uniform(1 - amount, 1 + amount)
                x_scaled[:, :, i] *= random_amount
    else:
        for i in prange(X.shape[0]):
            if additive:
                random_amount = np.random.uniform(-amount, amount)
                x_scaled[i, :, :] += random_amount
            else:
                random_amount = np.random.uniform(1 - amount, 1 + amount)
                x_scaled[i, :, :] *= random_amount

    return fit_data_to_dtype(x_scaled, X.dtype), y_scaled


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def augmentation_contrast(
    X: np.ndarray,
    y: Optional[np.ndarray],
    chance: float = 0.2,
    max_amount: float = 0.1,
    channel_last: bool = True,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Changes the contrast of an image by a random amount, seperately for each channel.
    input should be (height, width, channels) or (channels, height, width).

    Args:
        X (np.ndarray): The image to change the contrast of.
        y (np.ndarray/None): The label to change the contrast of. If None, no label is returned.

    Keyword Args:
        chance (float=0.2): The chance of changing the contrast.
        max_amount (float=0.1): The max amount to change the contrast by.
        channel_last (bool=True): Whether the image is (channels, height, width) or (height, width, channels).

    Returns:
        Tuple[np.ndarray, Optional[np.ndarray]]: The image with changed contrast and optionally the unmodified label.
    """
    if np.random.rand() > chance:
        return X, y

    x_contrast = X.astype(np.float32)
    y_contrast = y

    amount = np.random.rand() * max_amount

    channels = X.shape[2] if channel_last else X.shape[0]
    mean_pixel = np.array([0.0] * channels, dtype=np.float32)

    # Numba workaround
    if channel_last:
        for i in prange(channels):
            mean_pixel[i] = np.mean(x_contrast[:, :, i])
    else:
        for i in prange(channels):
            mean_pixel[i] = np.mean(x_contrast[i])

    for i in prange(channels):
        x_contrast[:, :, i] = (x_contrast[:, :, i] - mean_pixel[i]) * (1 + amount) + mean_pixel[i]

    return fit_data_to_dtype(x_contrast, X.dtype), y_contrast


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def augmentation_drop_pixel(
    X: np.ndarray,
    y: Optional[np.ndarray],
    chance: float = 0.2,
    drop_probability: float = 0.01,
    drop_value: float = 0.0,
    channel_last: bool = True,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Drops a random pixels from an image.
    input should be (height, width, channels) or (channels, height, width).
    Only drops pixels from features, not labels.

    Args:
        X (np.ndarray): The image to drop a pixel from.
        y (np.ndarray/None): The label to drop a pixel from. If None, no label is returned.

    Keyword Args:
        chance (float=0.2): The chance of dropping a pixel.
        drop_probability (float=0.05): The probability of dropping a pixel.
        drop_value (float=0.0): The value to drop the pixel to.
        apply_to_y (bool=False): Whether to apply the drop to the label.
        channel_last (bool=True): Whether the image is (channels, height, width) or (height, width, channels).

    Returns:
        Tuple[np.ndarray, Optional[np.ndarray]]: The image with the dropped pixels and optionally the unmodified label.
    """
    if np.random.rand() > chance:
        return X, y

    x_dropped = X.copy()
    y_dropped = y

    mask = np.random.random(size=x_dropped.shape)

    # Agreed. This looks terrible. But it's the only way to get numba to parallelize this.
    if channel_last:
        for col in prange(X.shape[0]):
            for row in prange(X.shape[1]):
                for channel in prange(X.shape[2]):
                    if mask[col, row, channel] <= drop_probability:
                        x_dropped[col, row, channel] = drop_value

    else:
        for channel in prange(X.shape[0]):
            for col in prange(X.shape[1]):
                for row in prange(X.shape[2]):
                    if mask[channel, col, row] <= drop_probability:
                        x_dropped[channel, col, row] = drop_value

    return x_dropped, y_dropped


@jit(nopython=True, nogil=True, cache=True, fastmath=True)
def augmentation_drop_channel(
    X: np.ndarray,
    y: Optional[np.ndarray],
    chance: float = 0.2,
    drop_probability: float = 0.1,
    drop_value: float = 0.0,
    channel_last: bool = True,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Drops a random channel from an image.
    input should be (height, width, channels) or (channels, height, width).
    A maximum of one channel will be dropped.

    Args:
        X (np.ndarray): The image to drop a channel from.
        y (np.ndarray/None): The label to drop a channel from. If None, no label is returned.

    Keyword Args:
        chance (float=0.2): The chance of dropping a channel.
        drop_probability (float=0.1): The probability of dropping a channel.
        drop_value (float=0.0): The value to drop the channel to.
        channel_last (bool=True): Whether the image is (channels, height, width) or (height, width, channels).
    """
    if np.random.rand() > chance:
        return X, y

    x_dropped = X.copy()
    y_dropped = y

    channels = X.shape[2] if channel_last else X.shape[0]

    drop_a_channel = False
    for _ in range(channels):
        if np.random.rand() < drop_probability:
            drop_a_channel = True
            break

    if not drop_a_channel:
        return X, y

    channel_to_drop = np.random.randint(0, channels)

    if channel_last:
        x_dropped[:, :, channel_to_drop] = drop_value
    else:
        x_dropped[channel_to_drop, :, :] = drop_value

    return x_dropped, y_dropped


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def augmentation_blur(
    X: np.ndarray,
    y: Optional[np.ndarray],
    chance: float = 0.2,
    intensity: float = 1.0,
    channel_last: bool = True,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Blurs an image at random.
    input should be (height, width, channels) or (channels, height, width).
    Same goes for the label if apply_to_y is True.

    Args:
        X (np.ndarray): The image to potentially blur.
        y (np.ndarray/None): The label is ignored.

    Keyword Args:
        chance (float=0.2): The chance of blurring a pixel.
        intensity (float=1.0): The intensity of the blur. from 0.0 to 1.0.
        apply_to_y (bool=False): Whether to blur the label as well.
        channel_last (bool=True): Whether the image is (channels, height, width) or (height, width, channels).
    """
    if np.random.rand() > chance:
        return X, y

    x_blurred = X.astype(np.float32)
    y_blurred = y

    offsets, weights = simple_blur_kernel_2d_3x3()
    channels = X.shape[2] if channel_last else X.shape[0]

    if channel_last:
        for channel in prange(channels):
            x_blurred[:, :, channel] = convolve_array_simple(x_blurred[:, :, channel], offsets, weights, intensity)
    else:
        for channel in prange(channels):
            x_blurred[channel, :, :] = convolve_array_simple(x_blurred[channel, :, :], offsets, weights, intensity)

    x_blurred = fit_data_to_dtype(x_blurred, X.dtype)

    return x_blurred, y_blurred


@jit(nopython=True, nogil=True, cache=True, fastmath=True)
def augmentation_blur_xy(
    X: np.ndarray,
    y: np.ndarray,
    chance: float = 0.2,
    intensity: float = 1.0,
    channel_last: bool = True,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Blurs an image at random.
    input should be (height, width, channels) or (channels, height, width).
    Same goes for the label.

    Args:
        X (np.ndarray): The image to potentially blur.
        y (np.ndarray/None): The label to potentially blur. If None, no label is returned.

    Keyword Args:
        chance (float=0.2): The chance of blurring a pixel.
        intensity (float=1.0): The intensity of the blur. from 0.0 to 1.0.
        apply_to_y (bool=False): Whether to blur the label as well.
        channel_last (bool=True): Whether the image is (channels, height, width) or (height, width, channels).
    """
    if np.random.rand() > chance:
        return X, y

    x_blurred, _ = augmentation_blur(X, None, chance, intensity, channel_last)
    y_blurred, _ = augmentation_blur(y, None, chance, intensity, channel_last)

    return x_blurred, y_blurred


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def augmentation_sharpen(
    X: np.ndarray,
    y: Optional[np.ndarray],
    chance: float = 0.2,
    intensity: float = 1.0,
    channel_last: bool = True,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Sharpens an image at random.
    input should be (height, width, channels) or (channels, height, width).

    Args:
        X (np.ndarray): The image to potentially sharpen.
        y (np.ndarray/None): The label will be ignored. If None, no label is returned.

    Keyword Args:
        chance (float=0.2): The chance of sharpening a pixel.
        intensity (float=1.0): The intensity of the sharpening. from 0.0 to 1.0.
        apply_to_y (bool=False): Whether to sharpen the label as well.
        channel_last (bool=True): Whether the image is (channels, height, width) or (height, width, channels).
    """
    if np.random.rand() > chance:
        return X, y

    x_sharpened = X.astype(np.float32)
    y_sharpened = y

    offsets, weights = simple_unsharp_kernel_2d_3x3()
    channels = X.shape[2] if channel_last else X.shape[0]

    if channel_last:
        for channel in prange(channels):
            x_sharpened[:, :, channel] = convolve_array_simple(x_sharpened[:, :, channel], offsets, weights, intensity)
    else:
        for channel in prange(channels):
            x_sharpened[channel, :, :] = convolve_array_simple(x_sharpened[channel, :, :], offsets, weights, intensity)

    x_sharpened = fit_data_to_dtype(x_sharpened, X.dtype)

    return x_sharpened, y_sharpened



@jit(nopython=True, nogil=True, cache=True, fastmath=True)
def augmentation_sharpen_xy(
    X: np.ndarray,
    y: np.ndarray,
    chance: float = 0.2,
    intensity: float = 1.0,
    channel_last: bool = True,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Sharpens an image at random.
    input should be (height, width, channels) or (channels, height, width).

    Args:
        X (np.ndarray): The image to potentially sharpen.
        y (np.ndarray): The label to potentially sharpen.

    Keyword Args:
        chance (float=0.2): The chance of sharpening a pixel.
        intensity (float=1.0): The intensity of the sharpening. from 0.0 to 1.0.
        channel_last (bool=True): Whether the image is (channels, height, width) or (height, width, channels).
    """
    if np.random.rand() > chance:
        return X, y

    x_sharpened, _ = augmentation_sharpen(X, None, chance, intensity, channel_last)
    y_sharpened, _ = augmentation_sharpen(y, None, chance, intensity, channel_last)

    return x_sharpened, y_sharpened


@jit(nopython=True, nogil=True, cache=True, fastmath=True)
def augmentation_misalign(
    X: np.ndarray,
    y: Optional[np.ndarray],
    chance: float = 0.2,
    max_offset: float = 0.5,
    channel_last: bool = True,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Misaligns one channel in the image at random.
    input should be (height, width, channels) or (channels, height, width).

    Args:
        X (np.ndarray): The image to potentially misalign the channels of.
        y (np.ndarray/None): The label to potentially misalign the channels of. If None, no label is returned.

    Keyword Args:
        chance (float=0.2): The chance of misaligning the channels of a pixel.
        max_offset (float=0.5): The maximum offset to misalign the channels by.
        channel_last (bool=True): Whether the image is (channels, height, width) or (height, width, channels).
    """
    if np.random.rand() > chance:
        return X, y

    x_misaligned = X.astype(np.float32)
    y_misaligned = y.astype(np.float32) if y is not None else None

    offsets, weights = simple_shift_kernel_2d(
        min(np.random.rand(), max_offset),
        min(np.random.rand(), max_offset),
    )

    channels = X.shape[2] if channel_last else X.shape[0]
    channel_to_adjust = np.random.randint(0, channels)

    if channel_last:
        x_misaligned[:, :, channel_to_adjust] = convolve_array_simple(x_misaligned[:, :, channel_to_adjust], offsets, weights)
    else:
        x_misaligned[channel_to_adjust, :, :] = convolve_array_simple(x_misaligned[channel_to_adjust, :, :], offsets, weights)

    x_misaligned = fit_data_to_dtype(x_misaligned, X.dtype)
    y_misaligned = fit_data_to_dtype(y_misaligned, y.dtype) if y is not None else None

    return x_misaligned, y_misaligned


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def augmentation_cutmix(
    X_target: np.ndarray,
    y_target: np.ndarray,
    X_source: np.ndarray,
    y_source: np.ndarray,
    chance: float = 0.2,
    min_size: float = 0.333,
    max_size: float = 0.666,
    label_mix: int = 0,
    feather: bool = False,
    feather_dist: int = 3,
    channel_last: bool = True,
) -> Tuple[np.ndarray, np.ndarray,]:
    """
    Cutmixes two images.
    input should be (height, width, channels) or (channels, height, width).

    Args:
        X_target (np.ndarray): The image to transfer the cutmix to.
        y_target (np.ndarray): The label to transfer the cutmix to.
        X_source (np.ndarray): The image to cutmix from.
        y_source (np.ndarray): The label to cutmix from.

    Keyword Args:
        chance (float=0.2): The chance of cutmixing a pixel.
        min_size (float=0.333): The minimum size of the patch to cutmix. In percentage of the image width.
        max_size (float=0.666): The maximum size of the patch to cutmix. In percentage of the image width.
        label_mix (int=0): if
            0 - The labels will be mixed by the weights.\n
            1 - The target label will be used.\n
            2 - The source label will be used.\n
            3 - The max of the labels will be used.\n
            4 - The min of the labels will be used.\n
            5 - The max of the image with the highest weight will be used.\n
            6 - The min of the image with the highest weight will be used.\n
            7 - The sum of the labels will be used.\n
        feather (bool=False): Whether to feather the edges of the cutmix.
        feather_dist (int=3): The distance to feather the edges of the cutmix in pixels
        channel_last (bool=True): Whether the image is (channels, height, width) or (height, width, channels).
    """
    if np.random.rand() > chance:
        return X_target, y_target

    x_mixed = X_target.astype(np.float32)
    y_mixed = y_target.astype(np.float32)

    if channel_last:
        height, width, channels_x = x_mixed.shape
    else:
        channels_x, height, width = x_mixed.shape

    channels_y = y_mixed.shape[2] if channel_last else y_mixed.shape[0]

    patch_height = np.random.randint(int(height * min_size), int(height * max_size))
    patch_width = np.random.randint(int(width * min_size), int(width * max_size))

    if feather:
        patch_height += feather_dist * 2
        patch_width += feather_dist * 2

        patch_height = min(patch_height, height)
        patch_width = min(patch_width, width)

    x0 = np.random.randint(0, width - patch_width)
    y0 = np.random.randint(0, height - patch_height)
    x1 = x0 + patch_width
    y1 = y0 + patch_height

    if feather:
        bbox = np.array([x0, x1, y0, y1])
        feather_weight_target = feather_box_2d(x_mixed, bbox, feather_dist)
        feather_weight_source = 1 - feather_weight_target

        x0 = max(0, x0 - feather_dist)
        x1 = min(x1 + 1 + feather_dist, x_mixed.shape[1])
        y0 = max(0, y0 - feather_dist)
        y1 = min(y1 + 1 + feather_dist, x_mixed.shape[0])

        if channel_last: # Reshape and tile instead of slicing, because of Numba.
            for col in prange(y0, y1):
                for row in prange(x0, x1):
                    for channel_x in prange(channels_x):
                        x_mixed[col, row, channel_x] = (
                            x_mixed[col, row, channel_x] * feather_weight_target[col, row][0]
                            + X_source[col, row, channel_x] * feather_weight_source[col, row][0]
                        )

                    for channel_y in prange(channels_y):
                        if label_mix == 0:
                            y_mixed[col, row, channel_y] = (
                                y_mixed[col, row, channel_y] * feather_weight_target[col, row][0]
                                + y_source[col, row, channel_y] * feather_weight_source[col, row][0]
                            )
                        elif label_mix == 1:
                            y_mixed[col, row, channel_y] = y_mixed[col, row, channel_y]
                        elif label_mix == 2:
                            y_mixed[col, row, channel_y] = y_source[col, row, channel_y]
                        elif label_mix == 3:
                            y_mixed[col, row, channel_y] = max(
                                y_mixed[col, row, channel_y], y_source[col, row, channel_y]
                            )
                        elif label_mix == 4:
                            y_mixed[col, row, channel_y] = min(
                                y_mixed[col, row, channel_y], y_source[col, row, channel_y]
                            )
                        elif label_mix == 5:
                            if feather_weight_target[col, row][0] > feather_weight_source[col, row][0]:
                                y_mixed[col, row, channel_y] = y_mixed[col, row, channel_y]
                            else:
                                y_mixed[col, row, channel_y] = y_source[col, row, channel_y]
                        elif label_mix == 6:
                            if feather_weight_target[col, row][0] < feather_weight_source[col, row][0]:
                                y_mixed[col, row, channel_y] = y_mixed[col, row, channel_y]
                            else:
                                y_mixed[col, row, channel_y] = y_source[col, row, channel_y]
                        elif label_mix == 7:
                            y_mixed[col, row, channel_y] = (
                                y_mixed[col, row, channel_y] + y_source[col, row, channel_y]
                            )
        else:
            for col in prange(y0, y1):
                for row in prange(x0, x1):
                    for channel_x in prange(channels_x):
                        x_mixed[channel_x, col, row] = (
                            x_mixed[channel_x, col, row] * feather_weight_target[col, row][0]
                            + X_source[channel_x, col, row] * feather_weight_source[col, row][0]
                        )

                    for channel_y in prange(channels_y):
                        if label_mix == 0:
                            y_mixed[channel_y, col, row] = (
                                y_mixed[channel_y, col, row] * feather_weight_target[col, row][0]
                                + y_source[channel_y, col, row] * feather_weight_source[col, row][0]
                            )
                        elif label_mix == 1:
                            y_mixed[channel_y, col, row] = y_mixed[channel_y, col, row]
                        elif label_mix == 2:
                            y_mixed[channel_y, col, row] = y_source[channel_y, col, row]
                        elif label_mix == 3:
                            y_mixed[channel_y, col, row] = max(
                                y_mixed[channel_y, col, row], y_source[channel_y, col, row]
                            )
                        elif label_mix == 4:
                            y_mixed[channel_y, col, row] = min(
                                y_mixed[channel_y, col, row], y_source[channel_y, col, row]
                            )
                        elif label_mix == 5:
                            if feather_weight_target[col, row][0] > feather_weight_source[col, row][0]:
                                y_mixed[channel_y, col, row] = y_mixed[channel_y, col, row]
                            else:
                                y_mixed[channel_y, col, row] = y_source[channel_y, col, row]
                        elif label_mix == 6:
                            if feather_weight_target[col, row][0] < feather_weight_source[col, row][0]:
                                y_mixed[channel_y, col, row] = y_mixed[channel_y, col, row]
                            else:
                                y_mixed[channel_y, col, row] = y_source[channel_y, col, row]
                        elif label_mix == 7:
                            y_mixed[channel_y, col, row] = (
                                y_mixed[channel_y, col, row] + y_source[channel_y, col, row]
                            )

    else:
        if channel_last:
            x_mixed[y0:y1, x0:x1, :] = X_source[y0:y1, x0:x1, :]
            y_mixed[y0:y1, x0:x1, :] = y_source[y0:y1, x0:x1, :]

        else:
            x_mixed[:, y0:y1, x0:x1] = X_source[:, y0:y1, x0:x1]
            y_mixed[:, y0:y1, x0:x1] = y_source[:, y0:y1, x0:x1]

    x_mixed = fit_data_to_dtype(x_mixed, X_target.dtype)
    y_mixed = fit_data_to_dtype(y_mixed, y_target.dtype)

    return x_mixed, y_mixed


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def augmentation_mixup(
    X_target: np.ndarray,
    y_target: np.ndarray,
    X_source: np.ndarray,
    y_source: np.ndarray,
    min_size: float = 0.333,
    max_size: float = 0.666,
    label_mix: int = 0,
    chance: float = 0.2,
    channel_last: bool = True, # pylint: disable=unused-argument
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Mixups two images at random. This works by doing a linear intepolation between
    two images and then adding a random weight to each image.

    Mixup involves taking two images and blending them together by randomly interpolating
    their pixel values. More specifically, suppose we have two images x and x' with their
    corresponding labels y and y'. To generate a new training example, mixup takes a
    weighted sum of x and x', such that the resulting image x^* = λx + (1-λ)x',
    where λ is a randomly chosen interpolation coefficient. The label for the new image
    is also a weighted sum of y and y' based on the same interpolation coefficient.

    input should be (height, width, channels) or (channels, height, width).

    Args:
        X_target (np.ndarray): The image to transfer to.
        y_target (np.ndarray): The label to transfer to.
        X_source (np.ndarray): The image to transfer from.
        y_source (np.ndarray): The label to transfer from.

    Keyword Args:
        min_size (float=0.333): The minimum mixup coefficient.
        max_size (float=0.666): The maximum mixup coefficient.
        label_mix (int=0): if
            0 - The labels will be mixed by the weights.\n
            1 - The target label will be used.\n
            2 - The source label will be used.\n
            3 - The max of the labels will be used.\n
            4 - The min of the labels will be used.\n
            5 - The max of the image with the highest weight will be used.\n
            6 - The min of the image with the highest weight will be used.\n
            7 - The sum of the labels will be used.\n
        chance (float=0.2): The chance of mixuping a pixel.
        channel_last (bool=True): Whether the image is (channels, height, width) or (height, width, channels).
    """
    if np.random.rand() > chance:
        return X_target, y_target

    x_mixed_target = X_target.astype(np.float32)
    y_mixed_target = y_target.astype(np.float32)

    x_mixed_source = X_source.astype(np.float32)
    y_mixed_source = y_source.astype(np.float32)

    mixup_coeff = np.float32(
        min(np.random.uniform(min_size, max_size + 0.001), 1.0),
    )

    x_mixed_target = x_mixed_target * mixup_coeff + x_mixed_source * (np.float32(1.0) - mixup_coeff)

    if label_mix == 0:
        y_mixed_target = y_mixed_target * mixup_coeff + y_mixed_source * (np.float32(1.0) - mixup_coeff)
    elif label_mix == 1:
        y_mixed_target = y_mixed_target
    elif label_mix == 2:
        y_mixed_target = y_mixed_source
    elif label_mix == 3:
        y_mixed_target = np.maximum(y_mixed_target, y_mixed_source)
    elif label_mix == 4:
        y_mixed_target = np.minimum(y_mixed_target, y_mixed_source)
    elif label_mix == 5:
        y_mixed_target = y_mixed_target if mixup_coeff >= 0.5 else y_mixed_source
    elif label_mix == 6:
        y_mixed_target = y_mixed_target if mixup_coeff >= 0.5 else y_mixed_source
    elif label_mix == 7:
        y_mixed_target = y_mixed_target + y_mixed_source

    x_mixed_target = fit_data_to_dtype(x_mixed_target, X_target.dtype)
    y_mixed_target = fit_data_to_dtype(y_mixed_target, y_target.dtype)

    return x_mixed_target, y_mixed_target



""" Slope, aspect, hillshade, and other DEM functions. """

# Standard library
import sys; sys.path.append("../../")
from typing import Union
from uuid import uuid4

# External
from osgeo import gdal
import numpy as np

# Internal
from buteo.raster import raster_to_array, array_to_raster
from buteo.utils.core_utils import type_check
from buteo.utils.gdal_utils import default_creation_options, delete_if_in_memory


def raster_dem_to_slope(
    input_raster: Union[str, gdal.Dataset],
    output_raster: str,
    slope_format: str = "percent",
    z_factor: float = 1.0,
) -> str:
    """
    Slope in percent.
    
    Args:
        input_raster: Path to input raster or gdal.Dataset.
        output_raster: Path to output raster.

    Keyword Args:
        slope_format: "percent" or "degree".
        z_factor: Z factor for slope calculation.

    Returns:
        Path to output raster.
    """
    type_check(input_raster, [str, gdal.Dataset], "input_raster")
    type_check(output_raster, str, "output_raster")
    type_check(z_factor, [float, int], "z_factor")

    slope_options = gdal.DEMProcessingOptions(
        computeEdges=True,
        slopeFormat=slope_format,
        zFactor=z_factor,
        creationOptions=default_creation_options(),
    )

    ds = gdal.DEMProcessing(
        output_raster,
        input_raster,
        'slope',
        options=slope_options,
        format=format,
    )

    if ds is None:
        raise RuntimeError("Error: Could not calculate slope")

    ds.FlushCache()  # Write to disk.
    ds = None  # Close the dataset and release resources

    return output_raster


def raster_dem_to_aspect(
    input_raster: Union[str, gdal.Dataset],
    output_raster: str,
    zero_for_flat: bool = True,
) -> str:
    """
    Aspect in degrees.
    
    Args:
        input_raster: Path to input raster or gdal.Dataset.
        output_raster: Path to output raster.

    Keyword Args:
        zero_for_flat: If True, set aspect to 0 for flat areas.

    Returns:
        Path to output raster.
    """
    type_check(input_raster, [str, gdal.Dataset], "input_raster")
    type_check(output_raster, str, "output_raster")

    aspect_options = gdal.DEMProcessingOptions(
        computeEdges=True,
        creationOptions=default_creation_options(),
        zeroForFlat=zero_for_flat,
    )

    ds = gdal.DEMProcessing(
        output_raster,
        input_raster,
        'aspect',
        options=aspect_options,
        format=format,
    )

    if ds is None:
        raise RuntimeError("Error: Could not calculate aspect")

    ds.FlushCache()  # Write to disk.
    ds = None  # Close the dataset and release resources


def raster_dem_to_hillshade(
    input_raster: Union[str, gdal.Dataset],
    output_raster: str,
    z_factor: float = 1.0,
) -> str:
    """
    Hillshade in degrees.
    
    Args:
        input_raster: Path to input raster or gdal.Dataset.
        output_raster: Path to output raster.

    Keyword Args:
        z_factor: Z factor for hillshade calculation.
    
    Returns:
        Path to output raster.
    """
    type_check(input_raster, [str, gdal.Dataset], "input_raster")
    type_check(output_raster, str, "output_raster")
    type_check(z_factor, [float, int], "z_factor")

    hillshade_options = gdal.DEMProcessingOptions(
        computeEdges=True,
        creationOptions=default_creation_options(),
        zFactor=z_factor,
    )

    ds = gdal.DEMProcessing(
        output_raster,
        input_raster,
        'hillshade',
        options=hillshade_options,
        format=format,
    )

    if ds is None:
        raise RuntimeError("Error: Could not calculate hillshade")

    ds.FlushCache()  # Write to disk.
    ds = None  # Close the dataset and release resources


def raster_dem_to_orientation(
    input_raster: Union[str, gdal.Dataset],
    output_raster: str,
    include_height: bool = True,
    height_normalisation: bool = True,
    height_normalisation_value: float = 8849.0, # mt. everest
) -> str:
    """ 
    Normalised orientation.

    Args:
        input_raster: Path to input raster or gdal.Dataset.
        output_raster: Path to output raster.
    
    Keyword Args:
        include_height: If True, include height in orientation calculation.
        height_normalisation: If True, normalise height to 0-1.
        height_normalisation_value: Value to normalise height to.
    
    Returns:
        Path to output raster.
    """
    tmp_path = f"/vsimem/{uuid4().int}_temp_aspect.tif"
    tmp_slope = f"/vsimem/{uuid4().int}_temp_slope.tif"

    raster_dem_to_aspect(input_raster, tmp_path, zero_for_flat=True)
    raster_dem_to_slope(input_raster, tmp_slope, slope_format="degree")

    aspect = raster_to_array(tmp_path)
    slope = raster_to_array(tmp_slope)

    delete_if_in_memory(tmp_path)
    delete_if_in_memory(tmp_slope)

    if include_height:
        dem = raster_to_array(input_raster)

        if height_normalisation:
            dem_norm = np.zeros(dem.shape, dtype=np.float32)
            dem_norm = np.divide(dem, height_normalisation_value, out=dem_norm, where=dem != 0)
        else:
            dem_norm = dem

        dem = None

    bands = 4 if include_height else 3

    destination = np.zeros((aspect.shape[0], aspect.shape[1], bands), dtype=np.float32)

    aspect_norm = np.zeros(aspect.shape, dtype=np.float32)
    aspect_norm = np.divide(aspect, 360.0, out=aspect_norm, where=aspect != 0)

    encoded_sin = ((np.sin(2 * np.pi * aspect_norm) + 1)) / 2.0
    encoded_cos = ((np.cos(2 * np.pi * aspect_norm) + 1)) / 2.0
    encoded_slope = np.divide(slope, 90.0, where=slope != 0)

    destination[:, :, 0] = encoded_sin[:, :, 0]
    destination[:, :, 1] = encoded_cos[:, :, 0]
    destination[:, :, 2] = encoded_slope[:, :, 0]

    if include_height:
        destination[:, :, 3] = dem_norm[:, :, 0]

    array_to_raster(destination, reference=input_raster, out_path=output_raster)

    return output_raster
